# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Implementation of an image service that uses Glance as the backend."""

import copy
import inspect
import os
import re
import stat
import sys
import time

import cryptography
from cursive import exception as cursive_exception
from cursive import signature_utils
import glanceclient
import glanceclient.exc
from glanceclient.v2 import schemas
from keystoneauth1 import loading as ks_loading
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils
from oslo_utils import timeutils

from cyborg.common import exception
from cyborg.common import utils
import cyborg.conf
from cyborg import objects
from cyborg import service_auth


LOG = logging.getLogger(__name__)
CONF = cyborg.conf.CONF

_SESSION = None


def _session_and_auth(context):
    # Session is cached, but auth needs to be pulled from context each time.
    global _SESSION

    if not _SESSION:
        _SESSION = ks_loading.load_session_from_conf_options(
            CONF, cyborg.conf.glance.glance_group.name)

    auth = service_auth.get_auth_plugin(context)

    return _SESSION, auth


def _glanceclient_from_endpoint(context, endpoint, version):
    sess, auth = _session_and_auth(context)

    return glanceclient.Client(version, session=sess, auth=auth,
                               endpoint_override=endpoint,
                               global_request_id=context.global_id)


def generate_glance_url(context):
    """Return the glance url."""
    return get_api_server(context)


def _endpoint_from_image_ref(image_href):
    """Return the image_ref and guessed endpoint from an image url.

    :param image_href: href of an image
    :returns: a tuple of the form (image_id, endpoint_url)
    """
    parts = image_href.split('/')
    image_id = parts[-1]
    # the endpoint is everything in the url except the last 3 bits
    # which are version, 'images', and image_id
    endpoint = '/'.join(parts[:-3])
    return (image_id, endpoint)


def generate_identity_headers(context, status='Confirmed'):
    return {
        'X-Auth-Token': getattr(context, 'auth_token', None),
        'X-User-Id': getattr(context, 'user_id', None),
        'X-Tenant-Id': getattr(context, 'project_id', None),
        'X-Roles': ','.join(getattr(context, 'roles', [])),
        'X-Identity-Status': status,
    }


def get_api_server(context):
    """Get the service endpoint."""
    sess, auth = _session_and_auth(context)
    ksa_adap = utils.get_ksa_adapter(
        cyborg.conf.glance.DEFAULT_SERVICE_TYPE,
        ksa_auth=auth, ksa_session=sess,
        min_version='2.0', max_version='2.latest')
    endpoint = utils.get_endpoint(ksa_adap)
    if endpoint:
        # NOTE(mriedem): Due to python-glanceclient bug 1707995 we have
        # to massage the endpoint URL otherwise it won't work properly.
        # We can't use glanceclient.common.utils.strip_version because
        # of bug 1748009.
        endpoint = re.sub(r'/v\d+(\.\d+)?/?$', '/', endpoint)

    return endpoint


class GlanceClientWrapper(object):
    """Glance client wrapper class that implements retries."""

    def __init__(self, context=None, endpoint=None):
        version = 2
        if endpoint is not None:
            self.client = self._create_static_client(context,
                                                     endpoint,
                                                     version)
        else:
            self.client = None
        self.api_server = None

    def _create_static_client(self, context, endpoint, version):
        """Create a client that we'll use for every call."""
        self.api_server = str(endpoint)
        return _glanceclient_from_endpoint(context, endpoint, version)

    def _create_onetime_client(self, context, version):
        """Create a client that will be used for one call."""
        if self.api_server is None:
            self.api_server = get_api_server(context)
        return _glanceclient_from_endpoint(context, self.api_server, version)

    def call(self, context, version, method, *args, **kwargs):
        """Call a glance client method.  If we get a connection error,
        retry the request according to CONF.glance.num_retries.
        """
        retry_excs = (glanceclient.exc.ServiceUnavailable,
                      glanceclient.exc.InvalidEndpoint,
                      glanceclient.exc.CommunicationError)
        num_attempts = 1 + CONF.glance.num_retries

        for attempt in range(1, num_attempts + 1):
            client = self.client or self._create_onetime_client(context,
                                                                version)
            try:
                controller = getattr(client,
                                     kwargs.pop('controller', 'images'))
                result = getattr(controller, method)(*args, **kwargs)
                if inspect.isgenerator(result):
                    # Convert generator results to a list, so that we can
                    # catch any potential exceptions now and retry the call.
                    return list(result)
                return result
            except retry_excs as e:
                if attempt < num_attempts:
                    extra = "retrying"
                else:
                    extra = 'done trying'

                LOG.exception("Error contacting glance server "
                              "'%(server)s' for '%(method)s', "
                              "%(extra)s.",
                              {'server': self.api_server,
                               'method': method, 'extra': extra})
                if attempt == num_attempts:
                    raise exception.GlanceConnectionFailed(
                        server=str(self.api_server), reason=str(e))
                time.sleep(1)


class GlanceImageServiceV2(object):
    """Provides storage and retrieval of disk image objects within Glance."""

    def __init__(self, client=None):
        self._client = client or GlanceClientWrapper()

    @staticmethod
    def _safe_fsync(fh):
        """Performs os.fsync on a filehandle only if it is supported.

        fsync on a pipe, FIFO, or socket raises OSError with EINVAL.  This
        method discovers whether the target filehandle is one of these types
        and only performs fsync if it isn't.

        :param fh: Open filehandle (not a path or fileno) to maybe fsync.
        """
        fileno = fh.fileno()
        mode = os.fstat(fileno).st_mode
        # A pipe answers True to S_ISFIFO
        if not any(check(mode) for check in (stat.S_ISFIFO, stat.S_ISSOCK)):
            os.fsync(fileno)

    def download(self, context, image_id, data=None, dst_path=None):
        """Calls out to Glance for data and writes data."""
        try:
            image_chunks = self._client.call(context, 2, 'data', image_id)
        except Exception:
            _reraise_translated_image_exception(image_id)

        if image_chunks.wrapped is None:
            # None is a valid return value, but there's nothing we can do with
            # a image with no associated data
            raise exception.ImageUnacceptable(image_id=image_id,
                                              reason='Image has no \
                                              associated data')

        # Retrieve properties for verification of Glance image signature
        verifier = None
        if CONF.glance.verify_glance_signatures:
            image_meta_dict = self.show(context, image_id,
                                        include_locations=False)
            image_meta = objects.ImageMeta.from_dict(image_meta_dict)
            img_signature = image_meta.properties.get('img_signature')
            img_sig_hash_method = image_meta.properties.get(
                'img_signature_hash_method'
            )
            img_sig_cert_uuid = image_meta.properties.get(
                'img_signature_certificate_uuid'
            )
            img_sig_key_type = image_meta.properties.get(
                'img_signature_key_type'
            )
            try:
                verifier = signature_utils.get_verifier(
                    context=context,
                    img_signature_certificate_uuid=img_sig_cert_uuid,
                    img_signature_hash_method=img_sig_hash_method,
                    img_signature=img_signature,
                    img_signature_key_type=img_sig_key_type,
                )
            except cursive_exception.SignatureVerificationError:
                with excutils.save_and_reraise_exception():
                    LOG.error('Image signature verification failed '
                              'for image: %s', image_id)

        close_file = False
        if data is None and dst_path:
            data = open(dst_path, 'wb')
            close_file = True

        if data is None:

            # Perform image signature verification
            if verifier:
                try:
                    for chunk in image_chunks:
                        verifier.update(chunk)
                    verifier.verify()

                    LOG.info('Image signature verification succeeded '
                             'for image: %s', image_id)

                except cryptography.exceptions.InvalidSignature:
                    with excutils.save_and_reraise_exception():
                        LOG.error('Image signature verification failed '
                                  'for image: %s', image_id)
            return image_chunks
        else:
            try:
                for chunk in image_chunks:
                    if verifier:
                        verifier.update(chunk)
                    data.write(chunk)
                if verifier:
                    verifier.verify()
                    LOG.info('Image signature verification succeeded '
                             'for image %s', image_id)
            except cryptography.exceptions.InvalidSignature:
                data.truncate(0)
                with excutils.save_and_reraise_exception():
                    LOG.error('Image signature verification failed '
                              'for image: %s', image_id)
            except Exception as ex:
                with excutils.save_and_reraise_exception():
                    LOG.error("Error writing to %(path)s: %(exception)s",
                              {'path': dst_path, 'exception': ex})
            finally:
                if close_file:
                    # Ensure that the data is pushed all the way down to
                    # persistent storage. This ensures that in the event of a
                    # subsequent host crash we don't have running instances
                    # using a corrupt backing file.
                    data.flush()
                    self._safe_fsync(data)
                    data.close()


def _extract_query_params(params):
    _params = {}
    accepted_params = ('filters', 'marker', 'limit',
                       'page_size', 'sort_key', 'sort_dir')
    for param in accepted_params:
        if params.get(param):
            _params[param] = params.get(param)

    # ensure filters is a dict
    _params.setdefault('filters', {})
    # NOTE(vish): don't filter out private images
    _params['filters'].setdefault('is_public', 'none')

    return _params


def _extract_query_params_v2(params):
    _params = {}
    accepted_params = ('filters', 'marker', 'limit',
                       'page_size', 'sort_key', 'sort_dir')
    for param in accepted_params:
        if params.get(param):
            _params[param] = params.get(param)

    # ensure filters is a dict
    _params.setdefault('filters', {})
    # NOTE(vish): don't filter out private images
    _params['filters'].setdefault('is_public', 'none')

    # adopt filters to be accepted by glance v2 api
    filters = _params['filters']
    new_filters = {}

    for filter_ in filters:
        # remove 'property-' prefix from filters by custom properties
        if filter_.startswith('property-'):
            new_filters[filter_.lstrip('property-')] = filters[filter_]
        elif filter_ == 'changes-since':
            # convert old 'changes-since' into new 'updated_at' filter
            updated_at = 'gte:' + filters['changes-since']
            new_filters['updated_at'] = updated_at
        elif filter_ == 'is_public':
            # convert old 'is_public' flag into 'visibility' filter
            # omit the filter if is_public is None
            is_public = filters['is_public']
            if is_public.lower() in ('true', '1'):
                new_filters['visibility'] = 'public'
            elif is_public.lower() in ('false', '0'):
                new_filters['visibility'] = 'private'
        else:
            new_filters[filter_] = filters[filter_]

    _params['filters'] = new_filters

    return _params


def _is_image_available(context, image):
    """Check image availability.

    This check is needed in case cyborg and Glance are deployed
    without authentication turned on.
    """
    # The presence of an auth token implies this is an authenticated
    # request and we need not handle the noauth use-case.
    if hasattr(context, 'auth_token') and context.auth_token:
        return True

    def _is_image_public(image):
        # NOTE(jaypipes) V2 Glance API replaced the is_public attribute
        # with a visibility attribute. We do this here to prevent the
        # glanceclient for a V2 image model from throwing an
        # exception from warlock when trying to access an is_public
        # attribute.
        if hasattr(image, 'visibility'):
            return str(image.visibility).lower() == 'public'
        else:
            return image.is_public

    if context.is_admin or _is_image_public(image):
        return True

    properties = image.properties

    if context.project_id and ('owner_id' in properties):
        return str(properties['owner_id']) == str(context.project_id)

    if context.project_id and ('project_id' in properties):
        return str(properties['project_id']) == str(context.project_id)

    try:
        user_id = properties['user_id']
    except KeyError:
        return False

    return str(user_id) == str(context.user_id)


def _translate_to_glance(image_meta):
    image_meta = _convert_to_string(image_meta)
    image_meta = _remove_read_only(image_meta)
    image_meta = _convert_to_v2(image_meta)
    return image_meta


def _convert_to_v2(image_meta):
    output = {}
    for name, value in image_meta.items():
        if name == 'properties':
            for prop_name, prop_value in value.items():
                # if allow_additional_image_properties is disabled we can't
                # define kernel_id and ramdisk_id as None, so we have to omit
                # these properties if they are not set.
                if prop_name in ('kernel_id', 'ramdisk_id') and \
                   prop_value is not None and \
                   prop_value.strip().lower() in ('none', ''):
                    continue
                # in glance only string and None property values are allowed,
                # v1 client accepts any values and converts them to string,
                # v2 doesn't - so we have to take care of it.
                elif prop_value is None or isinstance(
                        prop_value, str):
                    output[prop_name] = prop_value
                else:
                    output[prop_name] = str(prop_value)

        elif name in ('min_ram', 'min_disk'):
            output[name] = int(value)
        elif name == 'is_public':
            output['visibility'] = 'public' if value else 'private'
        elif name in ('size', 'deleted'):
            continue
        else:
            output[name] = value
    return output


def _translate_from_glance(image, include_locations=False):
    image_meta = _extract_attributes_v2(
        image, include_locations=include_locations)

    image_meta = _convert_timestamps_to_datetimes(image_meta)
    image_meta = _convert_from_string(image_meta)
    return image_meta


def _convert_timestamps_to_datetimes(image_meta):
    """Returns image with timestamp fields converted to datetime objects."""
    for attr in ['created_at', 'updated_at', 'deleted_at']:
        if image_meta.get(attr):
            image_meta[attr] = timeutils.parse_isotime(image_meta[attr])
    return image_meta


# NOTE(bcwaldon): used to store non-string data in glance metadata
def _json_loads(properties, attr):
    prop = properties[attr]
    if isinstance(prop, str):
        properties[attr] = jsonutils.loads(prop)


def _json_dumps(properties, attr):
    prop = properties[attr]
    if not isinstance(prop, str):
        properties[attr] = jsonutils.dumps(prop)


_CONVERT_PROPS = ('block_device_mapping', 'mappings')


def _convert(method, metadata):
    metadata = copy.deepcopy(metadata)
    properties = metadata.get('properties')
    if properties:
        for attr in _CONVERT_PROPS:
            if attr in properties:
                method(properties, attr)

    return metadata


def _convert_from_string(metadata):
    return _convert(_json_loads, metadata)


def _convert_to_string(metadata):
    return _convert(_json_dumps, metadata)


def _extract_attributes(image, include_locations=False):
    # TODO(mfedosin): Remove this function once we move to glance V2
    # completely.
    # NOTE(hdd): If a key is not found, base.Resource.__getattr__() may perform
    # a get(), resulting in a useless request back to glance. This list is
    # therefore sorted, with dependent attributes as the end
    # 'deleted_at' depends on 'deleted'
    # 'checksum' depends on 'status' == 'active'
    IMAGE_ATTRIBUTES = ['size', 'disk_format', 'owner',
                        'container_format', 'status', 'id',
                        'name', 'created_at', 'updated_at',
                        'deleted', 'deleted_at', 'checksum',
                        'min_disk', 'min_ram', 'is_public',
                        'direct_url', 'locations']

    queued = getattr(image, 'status') == 'queued'
    queued_exclude_attrs = ['disk_format', 'container_format']
    include_locations_attrs = ['direct_url', 'locations']
    output = {}

    for attr in IMAGE_ATTRIBUTES:
        if attr == 'deleted_at' and not output['deleted']:
            output[attr] = None
        elif attr == 'checksum' and output['status'] != 'active':
            output[attr] = None
        # image may not have 'name' attr
        elif attr == 'name':
            output[attr] = getattr(image, attr, None)
        # NOTE(liusheng): queued image may not have these attributes and 'name'
        elif queued and attr in queued_exclude_attrs:
            output[attr] = getattr(image, attr, None)
        # NOTE(mriedem): Only get location attrs if including locations.
        elif attr in include_locations_attrs:
            if include_locations:
                output[attr] = getattr(image, attr, None)
        # NOTE(mdorman): 'size' attribute must not be 'None', so use 0 instead
        elif attr == 'size':
            # NOTE(mriedem): A snapshot image may not have the size attribute
            # set so default to 0.
            output[attr] = getattr(image, attr, 0) or 0
        else:
            # NOTE(xarses): Anything that is caught with the default value
            # will result in an additional lookup to glance for said attr.
            # Notable attributes that could have this issue:
            # disk_format, container_format, name, deleted, checksum
            output[attr] = getattr(image, attr, None)

    output['properties'] = getattr(image, 'properties', {})

    return output


def _extract_attributes_v2(image, include_locations=False):
    include_locations_attrs = ['direct_url', 'locations']
    omit_attrs = ['self', 'schema', 'protected', 'virtual_size', 'file',
                  'tags']
    raw_schema = image.schema
    schema = schemas.Schema(raw_schema)
    output = {'properties': {}, 'deleted': False, 'deleted_at': None,
              'disk_format': None, 'container_format': None, 'name': None,
              'checksum': None}
    for name, value in image.items():
        if (name in omit_attrs
                or name in include_locations_attrs and not include_locations):
            continue
        elif name == 'visibility':
            output['is_public'] = value == 'public'
        elif name == 'size' and value is None:
            output['size'] = 0
        elif schema.is_base_property(name):
            output[name] = value
        else:
            output['properties'][name] = value

    return output


def _remove_read_only(image_meta):
    IMAGE_ATTRIBUTES = ['status', 'updated_at', 'created_at', 'deleted_at']
    output = copy.deepcopy(image_meta)
    for attr in IMAGE_ATTRIBUTES:
        if attr in output:
            del output[attr]
    return output


def _reraise_translated_image_exception(image_id):
    """Transform the exception for the image but keep its traceback intact."""
    exc_type, exc_value, exc_trace = sys.exc_info()
    new_exc = _translate_image_exception(image_id, exc_value)
    raise new_exc.with_traceback(exc_trace)


def _reraise_translated_exception():
    """Transform the exception but keep its traceback intact."""
    exc_type, exc_value, exc_trace = sys.exc_info()
    new_exc = _translate_plain_exception(exc_value)
    raise new_exc.with_traceback(exc_trace)


def _translate_image_exception(image_id, exc_value):
    if isinstance(exc_value, (glanceclient.exc.Forbidden,
                  glanceclient.exc.Unauthorized)):
        return exception.ImageNotAuthorized(image_id=image_id)
    if isinstance(exc_value, glanceclient.exc.NotFound):
        return exception.ResourceNotFound(
            resource='Image',
            msg='with uuid=%s' % image_id)
    if isinstance(exc_value, glanceclient.exc.BadRequest):
        return exception.ImageBadRequest(image_id=image_id,
                                         response=str(exc_value))
    return exc_value


def _translate_plain_exception(exc_value):
    if isinstance(exc_value, (glanceclient.exc.Forbidden,
                  glanceclient.exc.Unauthorized)):
        return exception.Forbidden(str(exc_value))
    if isinstance(exc_value, glanceclient.exc.NotFound):
        return exception.NotFound(str(exc_value))
    if isinstance(exc_value, glanceclient.exc.BadRequest):
        return exception.Invalid(str(exc_value))
    return exc_value


def get_remote_image_service(context, image_href):
    """Create an image_service and parse the id from the given image_href.

    The image_href param can be an href of the form
    'http://example.com:9292/v1/images/b8b2c6f7-7345-4e2f-afa2-eedaba9cbbe3',
    or just an id such as 'b8b2c6f7-7345-4e2f-afa2-eedaba9cbbe3'. If the
    image_href is a standalone id, then the default image service is returned.

    :param image_href: href that describes the location of an image
    :returns: a tuple of the form (image_service, image_id)

    """
    # NOTE(bcwaldon): If image_href doesn't look like a URI, assume its a
    # standalone image ID
    if '/' not in str(image_href):
        image_service = get_default_image_service()
        return image_service, image_href

    try:
        (image_id, endpoint) = _endpoint_from_image_ref(image_href)
        glance_client = GlanceClientWrapper(context=context,
                                            endpoint=endpoint)
    except ValueError:
        raise exception.InvalidImageRef(image_href=image_href)

    image_service = GlanceImageServiceV2(client=glance_client)
    return image_service, image_id


def get_default_image_service():
    return GlanceImageServiceV2()


class UpdateGlanceImage(object):
    def __init__(self, context, image_id, metadata, stream):
        self.context = context
        self.image_id = image_id
        self.metadata = metadata
        self.image_stream = stream

    def start(self):
        image_service, image_id = (
            get_remote_image_service(self.context, self.image_id))
        image_service.update(self.context, image_id, self.metadata,
                             self.image_stream, purge_props=False)



# Copyright 2019 Intel, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cyborg import objects
from cyborg.tests.unit.db import base
from cyborg.tests.unit.db import utils


class TestAttachHandleObject(base.DbTestCase):

    def setUp(self):
        super(TestAttachHandleObject, self).setUp()
        self.fake_attach_handle = utils.get_test_attach_handle()

    def test_get(self):
        uuid = self.fake_attach_handle['uuid']
        with mock.patch.object(self.dbapi, 'attach_handle_get_by_uuid',
                               autospec=True) as mock_attach_handle_get:
            mock_attach_handle_get.return_value = self.fake_attach_handle
            attach_handle = objects.AttachHandle.get(self.context, uuid)
            mock_attach_handle_get.assert_called_once_with(self.context, uuid)
            self.assertEqual(self.context, attach_handle._context)

    def test_get_by_id(self):
        id = self.fake_attach_handle['id']
        with mock.patch.object(self.dbapi, 'attach_handle_get_by_id',
                               autospec=True) as mock_attach_handle_get:
            mock_attach_handle_get.return_value = self.fake_attach_handle
            attach_handle = objects.AttachHandle.get_by_id(self.context, id)
            mock_attach_handle_get.assert_called_once_with(self.context, id)
            self.assertEqual(self.context, attach_handle._context)

    @mock.patch.object(objects.AttachHandle, 'list')
    def test_get_attach_handle_by_deployable_id(self, mock_list):
        mock_list.return_value = [self.fake_attach_handle]
        deployable_id = self.fake_attach_handle['deployable_id']
        ah_filter = {'deployable_id': deployable_id}
        attach_handles = objects.AttachHandle.get_ah_list_by_deployable_id(
            self.context, deployable_id)
        mock_list.assert_called_once_with(self.context, ah_filter)
        self.assertEqual(deployable_id, attach_handles[0]['deployable_id'])

    @mock.patch.object(objects.AttachHandle, 'list')
    def test_get_attach_handle_by_depid_attachinfo(self, mock_list):
        mock_list.return_value = [self.fake_attach_handle]
        deployable_id = self.fake_attach_handle['deployable_id']
        attach_info = self.fake_attach_handle['attach_info']
        ah_filter = {'deployable_id': deployable_id,
                     'attach_info': attach_info}
        attach_handles = objects.AttachHandle.get_ah_by_depid_attachinfo(
            self.context, deployable_id, attach_info)
        mock_list.assert_called_once_with(self.context, ah_filter)
        self.assertEqual(attach_info, attach_handles['attach_info'])

        # test objects.AttachHandle.list() return []
        mock_list.return_value = []
        attach_handle = objects.AttachHandle.get_ah_by_depid_attachinfo(
            self.context, deployable_id, attach_info)
        self.assertIsNone(attach_handle)

    def test_list(self):
        with mock.patch.object(self.dbapi, 'attach_handle_list',
                               autospec=True) as mock_attach_handle_list:
            mock_attach_handle_list.return_value = [self.fake_attach_handle]
            attach_handles = objects.AttachHandle.list(self.context)
            self.assertEqual(1, mock_attach_handle_list.call_count)
            self.assertEqual(1, len(attach_handles))
            self.assertIsInstance(attach_handles[0], objects.AttachHandle)
            self.assertEqual(self.context, attach_handles[0]._context)

    def test_list_with_filter(self):
        with mock.patch.object(self.dbapi, 'attach_handle_get_by_filters',
                               autospec=True) as mock_ah_with_filter_list:
            mock_ah_with_filter_list.return_value = [self.fake_attach_handle]
            filters = {'limit': 1}
            attach_handles = objects.AttachHandle.list(self.context, filters)
            self.assertEqual(1, mock_ah_with_filter_list.call_count)
            self.assertEqual(1, len(attach_handles))
            self.assertIsInstance(attach_handles[0], objects.AttachHandle)
            mock_ah_with_filter_list.assert_called_once_with(
                self.context,
                {},
                sort_dir='desc',
                sort_key='created_at',
                limit=1,
                marker=None,
                )

    @mock.patch.object(objects.base.CyborgObject, "_from_db_object")
    def test_attach_handle_allocate(self, mock_from_db_obj):
        deployable_id = self.fake_attach_handle['deployable_id']
        with mock.patch.object(self.dbapi, 'attach_handle_allocate',
                               autospec=True) as mock_ah_allocate:
            mock_ah_allocate.return_value = self.fake_attach_handle
            objects.AttachHandle.allocate(self.context, deployable_id)
            mock_from_db_obj.assert_called_once_with(
                mock.ANY,
                self.fake_attach_handle)

    def test_attach_handle_deallocate(self):
        attach_handle_uuid = self.fake_attach_handle['uuid']
        with mock.patch.object(self.dbapi, 'attach_handle_update',
                               autospec=True) as mock_ah_update:
            ah_obj = objects.AttachHandle(**self.fake_attach_handle)
            ah_obj.deallocate(self.context)
            mock_ah_update.assert_called_once_with(
                self.context,
                attach_handle_uuid,
                {"in_use": False})

    def test_create(self):
        with mock.patch.object(self.dbapi, 'attach_handle_create',
                               autospec=True) as mock_attach_handle_create:
            mock_attach_handle_create.return_value = self.fake_attach_handle
            attach_handle = objects.AttachHandle(self.context,
                                                 **self.fake_attach_handle)
            attach_handle.create(self.context)
            mock_attach_handle_create.assert_called_once_with(
                self.context, self.fake_attach_handle)
            self.assertEqual(self.context, attach_handle._context)

    def test_destroy(self):
        uuid = self.fake_attach_handle['uuid']
        with mock.patch.object(self.dbapi, 'attach_handle_get_by_uuid',
                               autospec=True) as mock_attach_handle_get:
            mock_attach_handle_get.return_value = self.fake_attach_handle
            with mock.patch.object(self.dbapi, 'attach_handle_delete',
                                   autospec=True) as mock_attach_handle_delete:
                attach_handle = objects.AttachHandle.get(self.context, uuid)
                attach_handle.destroy(self.context)
                mock_attach_handle_delete.assert_called_once_with(self.context,
                                                                  uuid)
                self.assertEqual(self.context, attach_handle._context)

    def test_update(self):
        uuid = self.fake_attach_handle['uuid']
        with mock.patch.object(self.dbapi, 'attach_handle_get_by_uuid',
                               autospec=True) as mock_attach_handle_get:
            mock_attach_handle_get.return_value = self.fake_attach_handle
            with mock.patch.object(self.dbapi, 'attach_handle_update',
                                   autospec=True) as mock_attach_handle_update:
                fake = self.fake_attach_handle
                fake["attach_info"] = "new_attach_info"
                mock_attach_handle_update.return_value = fake
                attach_handle = objects.AttachHandle.get(self.context, uuid)
                attach_handle.attach_info = 'new_attach_info'
                attach_handle.save(self.context)
                mock_attach_handle_get.assert_called_once_with(self.context,
                                                               uuid)
                mock_attach_handle_update.assert_called_once_with(
                    self.context, uuid,
                    {'attach_info': 'new_attach_info'})
                self.assertEqual(self.context, attach_handle._context)



"""add_ssd_type

Revision ID: 4cc1d79978fc
Revises: 899cead40bc9
Create Date: 2021-02-15 16:02:58.856126

"""

# revision identifiers, used by Alembic.
revision = '4cc1d79978fc'
down_revision = '899cead40bc9'

from alembic import op
import sqlalchemy as sa


def upgrade():
    new_device_type = sa.Enum('GPU', 'FPGA', 'AICHIP', 'QAT', 'NIC', 'SSD',
                             name='device_type')
    op.alter_column('devices', 'type',
                    existing_type=new_device_type,
                    nullable=False)



# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cyborg.api.controllers.v2 import versions
from cyborg.tests.unit.api.controllers.v2 import base as v2_test


class TestAPI(v2_test.APITestV2):

    def setUp(self):
        super(TestAPI, self).setUp()
        self.headers = self.gen_headers(self.context)

    def test_get_api_v2(self):
        data = self.get_json('/', headers=self.headers)
        self.assertEqual(data['status'], "CURRENT")
        self.assertEqual(data['max_version'], versions._MAX_VERSION_STRING)
        self.assertEqual(data['id'], "v2.0")
        result = isinstance(data['links'], list)
        self.assertTrue(result)



# Copyright 2019 Intel, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cyborg import objects
from cyborg.tests.unit.db import base
from cyborg.tests.unit.db import utils

from oslo_serialization import jsonutils


class TestControlpathIDObject(base.DbTestCase):

    def setUp(self):
        super(TestControlpathIDObject, self).setUp()
        self.fake_control_path = utils.get_test_control_path()

    def test_get(self):
        uuid = self.fake_control_path['uuid']
        with mock.patch.object(self.dbapi, 'control_path_get_by_uuid',
                               autospec=True) as mock_control_path_get:
            mock_control_path_get.return_value = self.fake_control_path
            control_path = objects.ControlpathID.get(self.context, uuid)
            mock_control_path_get.assert_called_once_with(self.context, uuid)
            self.assertEqual(self.context, control_path._context)

    def test_get_set_cpid_info_using_obj(self):
        uuid = self.fake_control_path['uuid']
        with mock.patch.object(self.dbapi, 'control_path_get_by_uuid',
                               autospec=True) as mock_control_path_get:
            mock_control_path_get.return_value = self.fake_control_path
            # test cpid_info_obj loader
            control_path = objects.ControlpathID.get(self.context, uuid)
            self.assertEqual(jsonutils.loads(control_path.cpid_info),
                             control_path.cpid_info_obj)
            # test cpid_info_obj setter
            control_path.cpid_info_obj = {'bus': "fake"}
            self.assertEqual(control_path.cpid_info,
                             jsonutils.dumps(control_path.cpid_info_obj))

    def test_list(self):
        with mock.patch.object(self.dbapi, 'control_path_list',
                               autospec=True) as mock_control_path_list:
            mock_control_path_list.return_value = [self.fake_control_path]
            control_paths = objects.ControlpathID.list(self.context)
            self.assertEqual(1, mock_control_path_list.call_count)
            self.assertEqual(1, len(control_paths))
            self.assertIsInstance(control_paths[0], objects.ControlpathID)
            self.assertEqual(self.context, control_paths[0]._context)

    def test_create(self):
        with mock.patch.object(self.dbapi, 'control_path_create',
                               autospec=True) as mock_control_path_create:
            mock_control_path_create.return_value = self.fake_control_path
            control_path = objects.ControlpathID(self.context,
                                                 **self.fake_control_path)
            control_path.create(self.context)
            mock_control_path_create.assert_called_once_with(
                self.context, self.fake_control_path)
            self.assertEqual(self.context, control_path._context)

    def test_destroy(self):
        uuid = self.fake_control_path['uuid']
        with mock.patch.object(self.dbapi, 'control_path_get_by_uuid',
                               autospec=True) as mock_control_path_get:
            mock_control_path_get.return_value = self.fake_control_path
            with mock.patch.object(self.dbapi, 'control_path_delete',
                                   autospec=True) as mock_control_path_delete:
                control_path = objects.ControlpathID.get(self.context, uuid)
                control_path.destroy(self.context)
                mock_control_path_delete.assert_called_once_with(self.context,
                                                                 uuid)
                self.assertEqual(self.context, control_path._context)

    def test_update(self):
        uuid = self.fake_control_path['uuid']
        with mock.patch.object(self.dbapi, 'control_path_get_by_uuid',
                               autospec=True) as mock_control_path_get:
            mock_control_path_get.return_value = self.fake_control_path
            with mock.patch.object(self.dbapi, 'control_path_update',
                                   autospec=True) as mock_control_path_update:
                fake = self.fake_control_path
                fake["cpid_info"] = "new_cpid_info"
                mock_control_path_update.return_value = fake
                control_path = objects.ControlpathID.get(self.context, uuid)
                control_path.cpid_info = 'new_cpid_info'
                control_path.save(self.context)
                mock_control_path_get.assert_called_once_with(self.context,
                                                              uuid)
                mock_control_path_update.assert_called_once_with(
                    self.context, uuid,
                    {'cpid_info': 'new_cpid_info'})
                self.assertEqual(self.context, control_path._context)



# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from cyborg.accelerator.drivers.gpu.base import GPUDriver
from cyborg.accelerator.drivers.gpu.nvidia.driver import NVIDIAGPUDriver
from cyborg.tests import base


class TestGPUDriver(base.TestCase):
    def test_create(self):
        # NVIDIAGPUDriver.VENDOR == 'nvidia'
        GPUDriver.create(NVIDIAGPUDriver.VENDOR)
        self.assertRaises(LookupError, GPUDriver.create, "matrox")

    def test_discover(self):
        d = GPUDriver()
        self.assertRaises(NotImplementedError, d.discover)



# Copyright 2020 Inspur Technologies Co.,LTD.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cyborg.objects.driver_objects import driver_attach_handle
from cyborg.objects.driver_objects import driver_attribute
from cyborg.objects.driver_objects import driver_controlpath_id
from cyborg.objects.driver_objects import driver_deployable
from cyborg.objects.driver_objects import driver_device


def get_fake_driver_devices_as_dict():
    driver_device1 = {
        "vendor": "0xABCD",
        "vendor_board_info": "fake_vendor_info",
        "model": "miss model info",
        "type": "GPU",
        "std_board_info": "{'class': 'Fake class', 'device_id': '0xabcd'}",
        "stub": False,
        "controlpath_id": get_fake_driver_controlpath_objs()[0],
        "deployable_list": get_fake_driver_deployable_objs()[:1]
    }
    driver_device2 = {
        "vendor": "0xDCBA",
        "vendor_board_info": "fake_vendor_info",
        "model": "miss model info",
        "type": "GPU",
        "std_board_info": "{'class': 'Fake class', 'device_id': '0xdcba'}",
        "stub": False,
        "controlpath_id": get_fake_driver_controlpath_objs()[1],
        "deployable_list": get_fake_driver_deployable_objs()[1:]
    }
    return [driver_device1, driver_device2]


def get_fake_driver_devices_objs():
    devices_list = get_fake_driver_devices_as_dict()
    obj_devices = []
    for device in devices_list:
        obj_driver_device = driver_device.DriverDevice()
        for field in device.keys():
            obj_driver_device[field] = device[field]
        obj_devices.append(obj_driver_device)
    return obj_devices


def get_fake_driver_controlpath_as_dict():
    driver_controlpath1 = {
        "cpid_info": '{"bus": "af", "device":00, "domain":0000, "function":0}',
        "cpid_type": "PCI"
    }
    driver_controlpath2 = {
        "cpid_info": '{"bus": "db", "device":00, "domain":0000, "function":0}',
        "cpid_type": "PCI"
    }
    return [driver_controlpath1, driver_controlpath2]


def get_fake_driver_controlpath_objs():
    controlpath_list = get_fake_driver_controlpath_as_dict()
    obj_controlpaths = []
    for controlpath in controlpath_list:
        obj_driver_controlpath = driver_controlpath_id.DriverControlPathID()
        for field in controlpath.keys():
            obj_driver_controlpath[field] = controlpath[field]
        obj_controlpaths.append(obj_driver_controlpath)
    return obj_controlpaths


def get_fake_driver_depolyables_as_dict():
    driver_depolyable1 = {
        "attach_handle_list": get_fake_driver_attach_handle_objs()[:1],
        "attribute_list": get_fake_driver_attribute_objs(),
        "driver_name": "NVIDIA",
        "name": "Tesla V100_0000:af:00.0",
        "num_accelerators": 1,
    }
    driver_depolyable2 = {
        "attach_handle_list": get_fake_driver_attach_handle_objs()[1:],
        "attribute_list": get_fake_driver_attribute_objs(),
        "driver_name": "NVIDIA",
        "name": "Tesla V100_0000:db:00.0",
        "num_accelerators": 1,
    }
    return [driver_depolyable1, driver_depolyable2]


def get_fake_driver_deployable_objs():
    deployables_list = get_fake_driver_depolyables_as_dict()
    obj_deployables = []
    for deployable in deployables_list:
        obj_driver_deployable = driver_deployable.DriverDeployable()
        for field in deployable.keys():
            obj_driver_deployable[field] = deployable[field]
        obj_deployables.append(obj_driver_deployable)
    return obj_deployables


def get_fake_driver_attach_handles_as_dict():
    driver_attach_handle1 = {
        "attach_info":
            '{"bus": "af", "device":00, "domain":0000, "function":0}',
        "attach_type": "PCI",
        "in_use": 0
    }
    driver_attach_handle2 = {
        "attach_info":
            '{"bus": "db", "device":00, "domain":0000, "function":0}',
        "attach_type": "PCI",
        "in_use": 0
    }
    return [driver_attach_handle1, driver_attach_handle2]


def get_fake_driver_attach_handle_objs():
    attach_handles_list = get_fake_driver_attach_handles_as_dict()
    obj_attach_handles = []
    for attach_handle in attach_handles_list:
        obj_driver_attach_handle = driver_attach_handle.DriverAttachHandle()
        for field in attach_handle.keys():
            obj_driver_attach_handle[field] = attach_handle[field]
        obj_attach_handles.append(obj_driver_attach_handle)
    return obj_attach_handles


def get_fake_driver_attributes_as_dict():
    driver_attribute1 = {
        "key": "trait0",
        "value": "CUSTOM_GPU_NVIDIA"
    }
    driver_attribute2 = {
        "key": "trait1",
        "value": "CUSTOM_GPU_PRODUCT_ID_1DB6"
    }
    driver_attribute3 = {
        "key": "rc",
        "value": "PGPU"
    }
    return [driver_attribute1, driver_attribute2, driver_attribute3]


def get_fake_driver_attribute_objs():
    attributes_list = get_fake_driver_attributes_as_dict()
    obj_attributes = []
    for attribute in attributes_list:
        obj_driver_attribute = driver_attribute.DriverAttribute()
        for field in attribute.keys():
            obj_driver_attribute[field] = attribute[field]
        obj_attributes.append(obj_driver_attribute)
    return obj_attributes



# Copyright 2018 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils

from cyborg.common import exception
from cyborg import db as db_api

LOG = logging.getLogger(__name__)

quota_opts = [
    cfg.IntOpt('reservation_expire',
               default=86400,
               help='Number of seconds until a reservation expires'),
    cfg.IntOpt('until_refresh',
               default=0,
               help='Count of reservations until usage is refreshed'),
    cfg.StrOpt('quota_driver',
               default="cyborg.quota.DbQuotaDriver",
               help='Default driver to use for quota checks'),
    cfg.IntOpt('quota_fpgas',
               default=10,
               help='Total amount of fpga allowed per project'),
    cfg.IntOpt('quota_gpus',
               default=10,
               help='Total amount of storage allowed per project'),
    cfg.IntOpt('max_age',
               default=0,
               help='Number of seconds between subsequent usage refreshes')
    ]

CONF = cfg.CONF
CONF.register_opts(quota_opts)


class QuotaEngine(object):
    """Represent the set of recognized quotas."""

    def __init__(self, quota_driver_class=None):
        """Initialize a Quota object."""

        self._resources = {}
        self._driver = DbQuotaDriver()

    def register_resource(self, resource):
        """Register a resource."""
        self._resources[resource.name] = resource

    def register_resources(self, resources):
        """Register a list of resources."""
        for resource in resources:
            self.register_resource(resource)

    def reserve(self, context, deltas, expire=None, project_id=None):
        """Check quotas and reserve resources.

        For counting quotas--those quotas for which there is a usage
        synchronization function--this method checks quotas against
        current usage and the desired deltas.  The deltas are given as
        keyword arguments, and current usage and other reservations
        are factored into the quota check.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it does not have a usage
        synchronization function.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns a
        list of reservation UUIDs which were created.

        :param context: The request context, for access checks.
        :param expire: An optional parameter specifying an expiration
                       time for the reservations.  If it is a simple
                       number, it is interpreted as a number of
                       seconds and added to the current time; if it is
                       a datetime.timedelta object, it will also be
                       added to the current time.  A datetime.datetime
                       object will be interpreted as the absolute
                       expiration time.  If None is specified, the
                       default expiration time set by
                       --default-reservation-expire will be used (this
                       value will be treated as a number of seconds).
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's project.
        """
        if not project_id:
            project_id = context.project_id
        reservations = self._driver.reserve(context, self._resources, deltas,
                                            expire=expire,
                                            project_id=project_id)

        LOG.debug("Created reservations %s", reservations)

        return reservations

    def commit(self, context, reservations, project_id=None):
        """Commit reservations.

        :param context: The request context, for access checks.
        :param reservations: A list of the reservation UUIDs, as
                             returned by the reserve() method.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's project.
        """
        project_id = context.project_id
        try:
            self._driver.commit(context, reservations, project_id=project_id)
        except Exception:
            # NOTE(Vek): Ignoring exceptions here is safe, because the
            # usage resynchronization and the reservation expiration
            # mechanisms will resolve the issue.  The exception is
            # logged, however, because this is less than optimal.
            LOG.exception("Failed to commit reservations %s", reservations)

    def rollback(self, context, reservations, project_id=None):
        pass


class DbQuotaDriver(object):
    """Driver to perform check to enforcement of quotas.

    Also allows to obtain quota information.
    The default driver utilizes the local database.
    """
    dbapi = db_api.get_instance()

    def reserve(self, context, resources, deltas, expire=None,
                project_id=None):
        # Set up the reservation expiration
        if expire is None:
            expire = CONF.reservation_expire
        if isinstance(expire, int):
            expire = datetime.timedelta(seconds=expire)
        if isinstance(expire, datetime.timedelta):
            expire = timeutils.utcnow() + expire
        if not isinstance(expire, datetime.datetime):
            raise exception.InvalidReservationExpiration(expire=expire)

        # If project_id is None, then we use the project_id in context
        if project_id is None:
            project_id = context.project_id

        return self._reserve(context, resources, deltas, expire,
                             project_id)

    def _reserve(self, context, resources, deltas, expire, project_id):
        return self.dbapi.quota_reserve(context, resources, deltas, expire,
                                        CONF.until_refresh, CONF.max_age,
                                        project_id=project_id)

    def commit(self, context, reservations, project_id=None):
        """Commit reservations.

        :param context: The request context, for access checks.
        :param reservations: A list of the reservation UUIDs, as
                             returned by the reserve() method.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's project.
        """

        try:
            self.dbapi.reservation_commit(context, reservations,
                                          project_id=project_id)
        except Exception:
            # NOTE(Vek): Ignoring exceptions here is safe, because the
            # usage resynchronization and the reservation expiration
            # mechanisms will resolve the issue.  The exception is
            # logged, however, because this is less than optimal.
            LOG.exception("Failed to commit reservations %s", reservations)


QUOTAS = QuotaEngine()



# Copyright 2024 Inspur.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg PCI driver implementation.
"""
from oslo_log import log as logging
from oslo_serialization import jsonutils

from cyborg.accelerator.common import utils
from cyborg.accelerator.drivers.pci import utils as pci_utils
from cyborg.accelerator.drivers.pci import whitelist
from cyborg.common import constants
import cyborg.conf
from cyborg.objects.driver_objects import driver_attach_handle
from cyborg.objects.driver_objects import driver_attribute
from cyborg.objects.driver_objects import driver_controlpath_id
from cyborg.objects.driver_objects import driver_deployable
from cyborg.objects.driver_objects import driver_device

LOG = logging.getLogger(__name__)
CONF = cyborg.conf.CONF


def _get_traits(vendor_id, product_id):
    """Generate traits for PCIs.
    : param vendor_id: vendor_id of PCI, eg."10de"
    : param product_id: product_id of PCI, eg."1eb8".

    Example PGPU traits:
    {traits:["OWNER_CYBORG", "CUSTOM_PCI_1EB8"]}
    """
    vendor_name = pci_utils.VENDOR_MAPS.get(vendor_id).upper()
    traits = ["CUSTOM_PCI_" + vendor_name]
    # PCIE trait
    product_trait = "_".join(('CUSTOM_PCI_PRODUCT_ID', product_id.upper()))
    traits.append(product_trait)
    return {"traits": traits}


def _generate_attribute_list(pci):
    attr_list = []
    index = 0
    for k, v in pci.items():
        if k == "rc":
            driver_attr = driver_attribute.DriverAttribute()
            driver_attr.key, driver_attr.value = k, v
            attr_list.append(driver_attr)
        if k == "traits":
            values = pci.get(k, [])
            for val in values:
                driver_attr = driver_attribute.DriverAttribute(
                    key="trait" + str(index), value=val)
                index = index + 1
                attr_list.append(driver_attr)
    return attr_list


def _generate_attach_handle(pci):
    driver_ah = driver_attach_handle.DriverAttachHandle()
    driver_ah.in_use = False
    driver_ah.attach_type = constants.AH_TYPE_PCI
    driver_ah.attach_info = utils.pci_str_to_json(pci["devices"])
    return driver_ah


def _generate_dep_list(pci):
    dep_list = []
    driver_dep = driver_deployable.DriverDeployable()
    driver_dep.attribute_list = _generate_attribute_list(pci)
    driver_dep.attach_handle_list = []
    # NOTE(wangzhh): The name of deployable should be unique, its format is
    # under disscussion, may looks like
    # <ComputeNodeName>_<NumaNodeName>_<CyborgName>_<NumInHost>
    # NOTE(yumeng) Since Wallaby release, the deplpyable_name is named as
    # <Compute_hostname>_<Device_address>
    driver_dep.name = pci.get('hostname', '') + '_' + pci["devices"]
    driver_dep.driver_name = \
        pci_utils.VENDOR_MAPS.get(pci["vendor_id"]).upper()
    driver_dep.num_accelerators = 1
    driver_dep.attach_handle_list = [_generate_attach_handle(pci)]
    dep_list.append(driver_dep)
    return dep_list, driver_dep.num_accelerators


def _generate_controlpath_id(pci):
    driver_cpid = driver_controlpath_id.DriverControlPathID()
    driver_cpid.cpid_type = "PCI"
    driver_cpid.cpid_info = utils.pci_str_to_json(pci["devices"])
    return driver_cpid


def _generate_driver_device(pci):
    driver_device_obj = driver_device.DriverDevice()
    driver_device_obj.vendor = pci['vendor_id']
    driver_device_obj.model = pci['product_id']
    std_board_info = {'product_id': pci.get('product_id'),
                      'controller': pci.get('controller'),
                      }
    driver_device_obj.std_board_info = jsonutils.dumps(std_board_info)
    driver_device_obj.type = constants.DEVICE_GPU
    driver_device_obj.stub = pci.get('stub', False)
    driver_device_obj.controlpath_id = _generate_controlpath_id(pci)
    driver_device_obj.deployable_list, ais = _generate_dep_list(pci)
    driver_device_obj.vendor_board_info = pci.get('vendor_board_info',
                                                  "miss_vb_info")
    return driver_device_obj


def _discover_pcis():
    cyborg.conf.devices.register_dynamic_opts(CONF)
    # discover pci devices by "lspci"
    pci_list = []
    pcis = pci_utils.get_pci_devices()
    LOG.info('pcis:%s', pcis)
    # report trait,rc and generate driver object
    dev_filter = whitelist.Whitelist(CONF.pci.passthrough_whitelist)
    for pci in pcis:
        m = dev_filter.device_assignable(pci)
        if m:
            pci_dict = m.groupdict()
            # get hostname for deployable_name usage
            pci_dict['hostname'] = CONF.host
            pci_dict["rc"] = constants.RESOURCES["PCI"]
            traits = _get_traits(pci_dict["vendor_id"],
                                 pci_dict["product_id"])
            pci_dict.update(traits)
            pci_list.append(_generate_driver_device(pci_dict))
    LOG.info('pci_list:%s', pci_list)
    return pci_list


def discover():
    devs = _discover_pcis()
    return devs



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import os_resource_classes as orc
from oslo_serialization import jsonutils

from cyborg.accelerator.drivers.driver import GenericDriver
from cyborg.common import constants
from cyborg.conf import CONF
from cyborg.objects.driver_objects import driver_attach_handle
from cyborg.objects.driver_objects import driver_attribute
from cyborg.objects.driver_objects import driver_controlpath_id
from cyborg.objects.driver_objects import driver_deployable
from cyborg.objects.driver_objects import driver_device


class FakeDriver(GenericDriver):
    """Base class for Fake drivers.

       This is just a Fake drivers interface.
    """
    VENDOR = "fake"
    NUM_ACCELERATORS = 16

    def _generate_controlpath_id(self, pci):
        driver_cpid = driver_controlpath_id.DriverControlPathID()
        driver_cpid.cpid_type = "PCI"
        driver_cpid.cpid_info = pci["slot"]
        return driver_cpid

    def _generate_attach_handles(self, pci, num_accelerators):
        """Returns list of attach handles, with same bus# but
        differing in device/function numbers. Assumes
        NUM_ACCELERATORS <= 256, otherwise bus# has to change too.
        """
        # In PCI bus-device-function address, 1 device can have 8 functions.
        NUM_PCI_FN_PER_PCI_DEVICE = 8

        ah_list = []
        for fn in range(1, num_accelerators + 1):  # fn 0 is CPID
            driver_ah = driver_attach_handle.DriverAttachHandle()
            # The virt driver will ignore this type when attaching
            driver_ah.attach_type = constants.AH_TYPE_TEST_PCI
            driver_ah.in_use = False

            pci_slot_dict = jsonutils.loads(pci['slot'])
            pci_slot_dict['device'] = str(int(fn / NUM_PCI_FN_PER_PCI_DEVICE))
            pci_slot_dict['function'] = str(fn % NUM_PCI_FN_PER_PCI_DEVICE)
            pci_slot_str = jsonutils.dumps(pci_slot_dict)

            driver_ah.attach_info = pci_slot_str
            ah_list.append(driver_ah)
        return ah_list

    def _generate_attribute_list(self):
        attr_traits = driver_attribute.DriverAttribute()
        attr_traits.key = "traits1"
        attr_traits.value = "CUSTOM_FAKE_DEVICE"
        attr_rc = driver_attribute.DriverAttribute()
        attr_rc.key = "rc"
        attr_rc.value = orc.FPGA
        return [attr_traits, attr_rc]

    def _generate_dep_list(self, pci):
        driver_dep = driver_deployable.DriverDeployable()
        driver_dep.attach_handle_list = self._generate_attach_handles(
            pci, self.NUM_ACCELERATORS)
        # NOTE(sean-k-mooney): we need to prepend the host name to the
        # device name as this is used to generate the RP name and uuid in
        # the cyborg conductor when updating placement. As such this needs
        # to be unique per host to allow multi node testing with the fake
        # driver.
        name = "%s_%s" % (CONF.host, pci.get('device'))
        # Replace any non alphanumeric, hyphen or underscore character with
        # underscore to comply with placement RP name requirements
        driver_dep.name = re.sub(r"(?![a-zA-Z0-9_\-]).", "_", name)
        driver_dep.driver_name = 'fake'
        driver_dep.num_accelerators = self.NUM_ACCELERATORS
        driver_dep.attribute_list = self._generate_attribute_list()
        return [driver_dep]

    def discover(self):
        fpga_list = []
        pci_addr = '{"domain":"0000","bus":"0c","device":"00","function":"0"}'
        pci_dict = {
            'slot': pci_addr,                    # PCI slot address
            'device': 'FakeDevice',              # Name of the device
            'vendor_id': '0xABCD',                 # ID of the vendor
            'class': 'Fake class',               # Name of the class
            'device_id': '0xabcd'                 # ID of the device
        }
        device = driver_device.DriverDevice()
        device.vendor = pci_dict["vendor_id"]
        device.model = pci_dict.get('model', 'miss model info')
        std_board_info = {'device_id': pci_dict.get('device_id'),
                          'class': pci_dict.get('class')}
        device.std_board_info = jsonutils.dumps(std_board_info)
        device.vendor_board_info = 'fake_vendor_info'
        device.type = constants.DEVICE_FPGA
        device.stub = False
        device.controlpath_id = self._generate_controlpath_id(pci_dict)
        device.deployable_list = self._generate_dep_list(pci_dict)
        fpga_list.append(device)
        return fpga_list

    def update(self, control_path, image_path):
        return True

    def get_stats(self):
        return {}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_concurrency import processutils
from oslo_serialization import jsonutils

import re

from cyborg.accelerator.common import utils
from cyborg.accelerator.drivers.driver import GenericDriver
from cyborg.common import constants
from cyborg.objects.driver_objects import driver_attach_handle
from cyborg.objects.driver_objects import driver_controlpath_id
from cyborg.objects.driver_objects import driver_deployable
from cyborg.objects.driver_objects import driver_device
import cyborg.privsep

PCI_INFO_PATTERN = re.compile(r"(?P<slot>[0-9a-f]{4}:[0-9a-f]{2}:"
                              r"[0-9a-f]{2}\.[0-9a-f]) "
                              r"(?P<class>.*) [\[].*]: (?P<device>.*) .*"
                              r"[\[](?P<vendor_id>[0-9a-fA-F]"
                              r"{4}):(?P<device_id>[0-9a-fA-F]{4})].*"
                              r"[(rev ](?P<revision>[0-9a-f]{2})")


@cyborg.privsep.sys_admin_pctxt.entrypoint
def lspci_privileged():
    cmd = ['lspci', '-nnn', '-D']
    return processutils.execute(*cmd)


class AscendDriver(GenericDriver):
    """The class for Ascend AI Chip drivers.

       This is the Huawei Ascend AI Chip drivers.
    """
    VENDOR = "huawei"

    # TODO(yikun): can be extracted into PCIDeviceDriver
    def _generate_controlpath_id(self, pci):
        driver_cpid = driver_controlpath_id.DriverControlPathID()
        driver_cpid.cpid_type = "PCI"
        driver_cpid.cpid_info = pci["slot_json"]
        return driver_cpid

    # TODO(yikun): can be extracted into PCIDeviceDriver
    def _generate_attach_handle(self, pci):
        driver_ah = driver_attach_handle.DriverAttachHandle()
        driver_ah.attach_type = "PCI"
        driver_ah.in_use = False
        driver_ah.attach_info = pci["slot_json"]
        return driver_ah

    # TODO(yikun): can be extracted into PCIDeviceDriver
    def _generate_dep_list(self, pci):
        driver_dep = driver_deployable.DriverDeployable()
        driver_dep.attach_handle_list = [self._generate_attach_handle(pci)]
        pci_addr_name = pci["slot"].replace(":", "_").replace(".", "_")
        driver_dep.name = pci.get('device', '') + '_' + pci_addr_name
        driver_dep.num_accelerators = 1
        driver_dep.driver_name = self.VENDOR
        return [driver_dep]

    # TODO(yikun): can be extracted into PCIDeviceDriver
    def _get_pci_lines(self, keywords=()):
        pci_lines = []
        if keywords:
            lspci_out = lspci_privileged()[0].split('\n')
            for i in range(len(lspci_out)):
                # filter out pci devices info that contains all keywords
                if all([k in (lspci_out[i]) for k in keywords]):
                    pci_lines.append(lspci_out[i])
        return pci_lines

    def discover(self):
        """The PCI line would be matched as:

        0000:0c:00.0 Processing acc [1200]: Device [19e5:d100] (rev 20)

        {
          'slot': '0000:0c:00.0',              # domain:bus:device.function
          'device': 'Device',                  # Name of the device
          'vendor_id': '19e5',                 # ID of the vendor
          'class': 'Processing accelerators',  # Name of the class
          'device_id': 'd100',                 # ID of the device
          'revision': '20'                     # Revision number
        }
        """
        ascends = self._get_pci_lines(('d100',))
        npu_list = []
        for ascend in ascends:
            m = PCI_INFO_PATTERN.match(ascend)
            if m:
                pci_dict = m.groupdict()
                pci_dict["slot_json"] = utils.pci_str_to_json(pci_dict["slot"])
                device = driver_device.DriverDevice()
                device.stub = False
                device.vendor = pci_dict["vendor_id"]
                device.model = pci_dict.get('model', '')
                std_board_info = {'device_id': pci_dict.get('device_id', None),
                                  'class': pci_dict.get('class', None)}
                device.std_board_info = jsonutils.dumps(std_board_info)
                device.vendor_board_info = ''
                device.type = constants.DEVICE_AICHIP
                device.controlpath_id = self._generate_controlpath_id(pci_dict)
                device.deployable_list = self._generate_dep_list(pci_dict)
                npu_list.append(device)
        return npu_list

    def update(self, control_path, image_path):
        # TODO(yikun): To be implemented in future
        return True

    def get_stats(self):
        # TODO(yikun): To be implemented in future
        return {}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
from oslo_serialization import jsonutils
from unittest import mock

from cyborg.common import exception
from cyborg.tests.unit.api.controllers.v2 import base as v2_test
from cyborg.tests.unit import fake_deployable
from cyborg.tests.unit import fake_device


class TestFPGAProgramController(v2_test.APITestV2):

    def setUp(self):
        super(TestFPGAProgramController, self).setUp()
        self.headers = self.gen_headers(self.context)
        self.deployable_uuids = ['0acbf8d6-e02a-4394-aae3-57557d209498']
        self.existent_image_uuid = "9a17439a-85d0-4c53-a3d3-0f68a2eac896"
        self.nonexistent_image_uuid = "1234abcd-1234-1234-1234-abcde1234567"
        self.invalid_image_uuid = "abcd1234"
        dep_uuid = self.deployable_uuids[0]
        self.dep = fake_deployable.fake_deployable_obj(self.context,
                                                       uuid=dep_uuid)
        self.dev = fake_device.get_fake_devices_objs()[0]
        bdf = {"domain": "0000", "bus": "00", "device": "01", "function": "1"}
        self.cpid = {
            "id": 0,
            "uuid": "e4a66b0d-b377-40d6-9cdc-6bf7e720e596",
            "device_id": "1",
            "cpid_type": "PCI",
            "cpid_info": jsonutils.dumps(bdf).encode('utf-8')
        }

    @mock.patch('cyborg.objects.Device.get_by_device_id')
    @mock.patch('cyborg.objects.Deployable.get_cpid_list')
    @mock.patch('cyborg.objects.Deployable.get')
    @mock.patch('cyborg.agent.rpcapi.AgentAPI.fpga_program')
    def test_program_success(self, mock_program, mock_get_dep,
                             mock_get_cpid_list, mock_get_by_device_id):
        self.headers['X-Roles'] = 'admin'
        self.headers['Content-Type'] = 'application/json'
        dep_uuid = self.deployable_uuids[0]
        mock_get_dep.return_value = self.dep
        mock_get_by_device_id.return_value = self.dev
        mock_get_cpid_list.return_value = [self.cpid]
        mock_program.return_value = True
        body = [{"image_uuid": self.existent_image_uuid}]
        response = self.patch_json('/deployables/%s/program' % dep_uuid,
                                   [{'path': '/bitstream_id', 'value': body,
                                     'op': 'replace'}], headers=self.headers)
        self.assertEqual(HTTPStatus.OK, response.status_code)
        data = response.json_body
        self.assertEqual(dep_uuid, data['uuid'])

    @mock.patch('cyborg.objects.Device.get_by_device_id')
    @mock.patch('cyborg.objects.Deployable.get_cpid_list')
    @mock.patch('cyborg.objects.Deployable.get')
    @mock.patch('cyborg.agent.rpcapi.AgentAPI.fpga_program')
    def test_program_failed(self, mock_program, mock_get_dep,
                            mock_get_cpid_list, mock_get_by_device_id):
        self.headers['X-Roles'] = 'admin'
        self.headers['Content-Type'] = 'application/json'
        dep_uuid = self.deployable_uuids[0]
        mock_get_dep.return_value = self.dep
        mock_get_by_device_id.return_value = self.dev
        mock_get_cpid_list.return_value = [self.cpid]
        mock_program.return_value = False
        body = [{"image_uuid": self.existent_image_uuid}]
        try:
            self.patch_json('/deployables/%s/program' % dep_uuid,
                            [{'path': '/bitstream_id', 'value': body,
                              'op': 'replace'}], headers=self.headers)
        except Exception as e:
            exc = e
        self.assertIn(exception.FPGAProgramError(
                      ret=mock_program.return_value).args[0],
                      exc.args[0]
                      )

    @mock.patch('cyborg.objects.Device.get_by_device_id')
    @mock.patch('cyborg.objects.Deployable.get_cpid_list')
    @mock.patch('cyborg.objects.Deployable.get')
    @mock.patch('cyborg.agent.rpcapi.AgentAPI.fpga_program')
    def test_program_invalid_uuid(self, mock_program, mock_get_dep,
                                  mock_get_cpid_list, mock_get_by_device_id):
        self.headers['X-Roles'] = 'admin'
        self.headers['Content-Type'] = 'application/json'
        dep_uuid = self.deployable_uuids[0]
        mock_get_dep.return_value = self.dep
        mock_get_by_device_id.return_value = self.dev
        mock_get_cpid_list.return_value = [self.cpid]
        mock_program.return_value = False
        body = [{"image_uuid": self.invalid_image_uuid}]
        try:
            self.patch_json('/deployables/%s/program' % dep_uuid,
                            [{'path': '/bitstream_id',
                              'value': body,
                              'op': 'replace'}],
                            headers=self.headers)
        except Exception as e:
            exc = e
        self.assertIn(exception.InvalidUUID(self.invalid_image_uuid).args[0],
                      exc.args[0])

    @mock.patch('cyborg.objects.Device.get_by_device_id')
    @mock.patch('cyborg.objects.Deployable.get_cpid_list')
    @mock.patch('cyborg.objects.Deployable.get')
    @mock.patch('cyborg.agent.rpcapi.AgentAPI.fpga_program')
    def test_program_wrong_image_uuid(self, mock_program,
                                      mock_get_dep,
                                      mock_get_cpid_list,
                                      mock_get_by_device_id):
        self.headers['X-Roles'] = 'admin'
        self.headers['Content-Type'] = 'application/json'
        dep_uuid = self.deployable_uuids[0]
        mock_get_dep.return_value = self.dep
        mock_get_by_device_id.return_value = self.dev
        mock_get_cpid_list.return_value = [self.cpid]
        mock_program.return_value = False
        body = [{"image_uuid": self.nonexistent_image_uuid}]
        try:
            self.patch_json('/deployables/%s/program' % dep_uuid,
                            [{'path': '/bitstream_id',
                              'value': body,
                              'op': 'replace'}],
                            headers=self.headers)
        except Exception as e:
            exc = e
        self.assertIn(exception.FPGAProgramError(
                      ret=mock_program.return_value).args[0],
                      exc.args[0]
                      )



# Copyright 2018 Beijing Lenovo Software Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import collections
import importlib
import os
import pkgutil

LIST_OPTS_FUNC_NAME = "list_opts"


def _tupleize(dct):
    """Take the dict of options and convert to the 2-tuple format."""
    return [(key, val) for key, val in dct.items()]


def list_opts():
    opts = collections.defaultdict(list)
    module_names = _list_module_names()
    imported_modules = _import_modules(module_names)
    _append_config_options(imported_modules, opts)
    return _tupleize(opts)


def _list_module_names():
    module_names = []
    package_path = os.path.dirname(os.path.abspath(__file__))
    for _, modname, ispkg in pkgutil.iter_modules(path=[package_path]):
        if modname == "opts" or ispkg:
            continue
        else:
            module_names.append(modname)
    return module_names


def _import_modules(module_names):
    imported_modules = []
    for modname in module_names:
        mod = importlib.import_module("cyborg.conf." + modname)
        if not hasattr(mod, LIST_OPTS_FUNC_NAME):
            msg = "The module 'zun.conf.%s' should have a '%s' "\
                  "function which returns the config options." % \
                  (modname, LIST_OPTS_FUNC_NAME)
            raise AttributeError(msg)
        else:
            imported_modules.append(mod)
    return imported_modules


def _append_config_options(imported_modules, config_options):
    for mod in imported_modules:
        configs = mod.list_opts()
        for key, val in configs.items():
            config_options[key].extend(val)



# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_versionedobjects import base as object_base

from cyborg.common import constants
from cyborg.db import api as dbapi
from cyborg.objects import base
from cyborg.objects import fields as object_fields

LOG = logging.getLogger(__name__)


@base.CyborgObjectRegistry.register
class ControlpathID(base.CyborgObject, object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    # Version 1.1: Add cpid_info_obj
    VERSION = '1.1'

    dbapi = dbapi.get_instance()

    fields = {
        'id': object_fields.IntegerField(nullable=False),
        'uuid': object_fields.UUIDField(nullable=False),
        'device_id': object_fields.IntegerField(nullable=False),
        'cpid_type': object_fields.EnumField(
            valid_values=constants.CPID_TYPE,
            nullable=False),
        'cpid_info': object_fields.StringField(nullable=False)
    }

    @property
    def cpid_info_obj(self):
        return jsonutils.loads(self.cpid_info)

    @cpid_info_obj.setter
    def cpid_info_obj(self, cpid_info_obj):
        self.cpid_info = jsonutils.dumps(cpid_info_obj)

    def create(self, context):
        """Create a ControlPathID record in the DB."""
        values = self.obj_get_changes()
        db_cp = self.dbapi.control_path_create(context, values)
        self._from_db_object(self, db_cp)

    @classmethod
    def get(cls, context, uuid):
        """Find a DB ControlpathID and return an Obj ControlpathID."""
        db_cp = cls.dbapi.control_path_get_by_uuid(context, uuid)
        obj_cp = cls._from_db_object(cls(context), db_cp)
        return obj_cp

    @classmethod
    def list(cls, context, filters=None):
        """Return a list of ControlpathID objects."""
        if filters:
            sort_dir = filters.pop('sort_dir', 'desc')
            sort_key = filters.pop('sort_key', 'created_at')
            limit = filters.pop('limit', None)
            marker = filters.pop('marker_obj', None)
            db_cps = cls.dbapi.control_path_get_by_filters(context, filters,
                                                           sort_dir=sort_dir,
                                                           sort_key=sort_key,
                                                           limit=limit,
                                                           marker=marker)
        else:
            db_cps = cls.dbapi.control_path_list(context)
        obj_cp_list = cls._from_db_object_list(db_cps, context)
        return obj_cp_list

    def save(self, context):
        """Update an ControlpathID record in the DB"""
        updates = self.obj_get_changes()
        db_cps = self.dbapi.control_path_update(context, self.uuid, updates)
        self._from_db_object(self, db_cps)

    def destroy(self, context):
        """Delete a ControlpathID from the DB."""
        self.dbapi.control_path_delete(context, self.uuid)
        self.obj_reset_changes()

    @classmethod
    def get_by_device_id(cls, context, device_id):
        # control_path is unique for one device.
        cpid_filter = {'device_id': device_id}
        cpid_obj_list = ControlpathID.list(context, cpid_filter)
        if len(cpid_obj_list) != 0:
            return cpid_obj_list[0]
        else:
            return None

    @classmethod
    def get_by_device_id_cpidinfo(cls, context, device_id, cpid_info):
        cpid_filter = {'device_id': device_id,
                       'cpid_info': cpid_info}
        # the list could have one value or is empty.
        cpid_obj_list = ControlpathID.list(context, cpid_filter)
        if len(cpid_obj_list) != 0:
            return cpid_obj_list[0]
        else:
            return None



# Copyright 2019 Intel.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from io import BytesIO
import json
import requests
from requests import structures
from requests import utils
from unittest import mock

from testtools.matchers import HasLength

from cyborg.common import constants
from cyborg.common import exception
from cyborg import objects
from cyborg.tests.unit.db import base
from cyborg.tests.unit import fake_deployable
from cyborg.tests.unit import fake_extarq


class TestFPGAExtARQObject(base.DbTestCase):

    def setUp(self):
        super(TestFPGAExtARQObject, self).setUp()
        self.fake_fpga_db_extarqs = fake_extarq.get_fake_fpga_db_extarqs()
        self.fake_obj_fpga_extarqs = fake_extarq.get_fake_fpga_extarq_objs()
        classes = ["no_program", "bitstream_program",
                   "function_program", "bad_program"]
        self.class_fgpa_objects = dict(
            zip(classes, self.fake_obj_fpga_extarqs))
        self.class_fgpa_dbs = dict(
            zip(classes, self.fake_fpga_db_extarqs))
        self.bitstream_id = self.class_fgpa_objects["bitstream_program"][
            "device_profile_group"][constants.ACCEL_BITSTREAM_ID]
        self.function_id = self.class_fgpa_objects["function_program"][
            "device_profile_group"][constants.ACCEL_FUNCTION_ID]
        self.images_md = {
            "/images":
            [
                {"id": self.bitstream_id,
                 "tags": ["trait:CUSTOM_FPGA_INTEL"],
                 constants.ACCEL_FUNCTION_ID: self.function_id},
            ]
        }
        self.deployable_uuids = ['0acbf8d6-e02a-4394-aae3-57557d209498']
        self.bdf = {"domain": "0000", "bus": "00",
                    "device": "01", "function": "1"}
        self.cpid = {
            "id": 0,
            "uuid": "e4a66b0d-b377-40d6-9cdc-6bf7e720e596",
            "device_id": "1",
            "cpid_type": "PCI",
            "cpid_info": json.dumps(self.bdf).encode('utf-8')
        }

    def response(self, status_code=200, content='', headers=None,
                 reason=None, elapsed=0, request=None, stream=False):
        res = requests.Response()
        res.status_code = status_code
        if isinstance(content, (dict, list)):
            content = json.dumps(content).encode('utf-8')
        if isinstance(content, str):
            content = content.encode('utf-8')
        res._content = content
        res._content_consumed = content
        res.headers = structures.CaseInsensitiveDict(headers or {})
        res.encoding = utils.get_encoding_from_headers(res.headers)
        res.reason = reason
        res.request = request
        if hasattr(request, 'url'):
            res.url = request.url
            if isinstance(request.url, bytes):
                res.url = request.url.decode('utf-8')

        if stream:
            res.raw = BytesIO(content)
        else:
            res.raw = BytesIO(b'')
        # normally this closes the underlying connection,
        #  but we have nothing to free.
        res.close = lambda *args, **kwargs: None
        return res

    def images_get(self, url, *args, **kw):
        images = {"images": []}
        params = kw.get("params", {})
        bit_id = None
        res = None
        if url != "/images":
            bit_id = url.rsplit("/", 1)[-1]
            params["id"] = bit_id
        tags = kw.pop("tags", [])
        for image in self.images_md["/images"]:
            if tags <= image["tags"] and params.items() <= image.items():
                images["images"].append(image)
                res = self.response(content=image)
        return res if bit_id else self.response(content=images)

    @mock.patch('cyborg.objects.ExtARQ._from_db_object')
    def test_get_bitstream_id_return_None(self, mock_from_db_obj):
        db_extarq = self.fake_fpga_db_extarqs[0]
        uuid = db_extarq['uuid']
        mock_from_db_obj.return_value = self.fake_obj_fpga_extarqs[0]
        with mock.patch.object(
            self.dbapi, 'extarq_get') as mock_extarq_get:
            mock_extarq_get.return_value = db_extarq
            obj_extarq = objects.FPGAExtARQ.get(self.context, uuid)
            mock_extarq_get.assert_called_once_with(self.context, uuid)
            self.assertEqual(obj_extarq.arq.uuid, uuid)
            self.assertIsNone(obj_extarq._get_bitstream_id())

    @mock.patch('cyborg.objects.ExtARQ._from_db_object')
    def test_get_bitstream_id_return_UUID(self, mock_from_db_obj):
        db_extarq = self.fake_fpga_db_extarqs[1]
        bit_id = db_extarq['device_profile_group'][
            constants.ACCEL_BITSTREAM_ID]
        uuid = db_extarq['uuid']
        mock_from_db_obj.return_value = self.fake_obj_fpga_extarqs[1]
        with mock.patch.object(
            self.dbapi, 'extarq_get') as mock_extarq_get:
            mock_extarq_get.return_value = db_extarq
            obj_extarq = objects.FPGAExtARQ.get(self.context, uuid)
            mock_extarq_get.assert_called_once_with(self.context, uuid)
            self.assertEqual(obj_extarq.arq.uuid, uuid)
            self.assertEqual(obj_extarq._get_bitstream_id(), bit_id)

    @mock.patch('cyborg.objects.ExtARQ._from_db_object')
    def test_get_function_id_return_UUID(self, mock_from_db_obj):
        db_extarq = self.fake_fpga_db_extarqs[2]
        fun_id = db_extarq['device_profile_group'][constants.ACCEL_FUNCTION_ID]
        uuid = db_extarq['uuid']
        mock_from_db_obj.return_value = self.fake_obj_fpga_extarqs[2]
        with mock.patch.object(
            self.dbapi, 'extarq_get') as mock_extarq_get:
            mock_extarq_get.return_value = db_extarq
            obj_extarq = objects.FPGAExtARQ.get(self.context, uuid)
            mock_extarq_get.assert_called_once_with(self.context, uuid)
            self.assertEqual(obj_extarq.arq.uuid, uuid)
            self.assertEqual(obj_extarq._get_function_id(), fun_id)

    @mock.patch('cyborg.objects.ExtARQ._from_db_object_list')
    def test_list(self, mock_from_db_obj_list):
        db_extarqs = self.fake_fpga_db_extarqs
        mock_from_db_obj_list.return_value = self.fake_obj_fpga_extarqs
        with mock.patch.object(self.dbapi, 'extarq_list',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = db_extarqs
            obj_extarqs = objects.FPGAExtARQ.list(self.context)
            self.assertEqual(1, mock_get_list.call_count)
            self.assertThat(obj_extarqs, HasLength(4))
            self.assertIsInstance(obj_extarqs[0], objects.FPGAExtARQ)
            for i, obj_extarq in enumerate(obj_extarqs):
                self.assertEqual(obj_extarq.arq.uuid, db_extarqs[i]['uuid'])

    @mock.patch('openstack.connection.Connection')
    def test_get_bitstream_md_from_bitstream_id(self, mock_conn):
        mock_conn.return_value = type(
            "Connection", (object,),
            {"image": type("image", (object,), {"get": self.images_get})})
        obj_extarq = self.class_fgpa_objects["bitstream_program"]
        md = obj_extarq._get_bitstream_md_from_bitstream_id(self.bitstream_id)
        self.assertDictEqual(self.images_md["/images"][0], md)

    @mock.patch('openstack.connection.Connection')
    def test_get_bitstream_md_from_function_id(self, mock_conn):
        mock_conn.return_value = type(
            "Connection", (object,),
            {"image": type("image", (object,), {"get": self.images_get})})
        obj_extarq = self.class_fgpa_objects["function_program"]
        md = obj_extarq._get_bitstream_md_from_function_id(self.function_id)
        self.assertDictEqual(self.images_md["/images"][0], md)

    @mock.patch('openstack.connection.Connection')
    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    def test_needs_programming(self, mock_check_state, mock_conn):
        mock_conn.return_value = type(
            "Connection", (object,),
            {"image": type("image", (object,), {"get": self.images_get})})
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(
            self.context, uuid=dep_uuid)

        # function_id is require
        obj_extarq = self.class_fgpa_objects["function_program"]
        need = obj_extarq._needs_programming(self.context, fake_dep)
        self.assertTrue(need)

        # bitstream_id is require
        obj_extarq = self.class_fgpa_objects["bitstream_program"]
        need = obj_extarq._needs_programming(self.context, fake_dep)
        self.assertTrue(need)

        # Both bitstream_id and function_id are require
        obj_extarq = self.class_fgpa_objects["bad_program"]
        self.assertRaises(
            exception.InvalidParameterValue, obj_extarq._needs_programming,
            self.context, fake_dep)

        # None of bitstream_id or function_id is require
        obj_extarq = self.class_fgpa_objects["no_program"]
        need = obj_extarq._needs_programming(self.context, fake_dep)
        self.assertFalse(need)

    @mock.patch('openstack.connection.Connection')
    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    def test_get_bitstream_md(self, mock_check_state, mock_conn):
        mock_conn.return_value = type(
            "Connection", (object,),
            {"image": type("image", (object,), {"get": self.images_get})})
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(
            self.context, uuid=dep_uuid)

        # function_id is require
        obj_extarq = self.class_fgpa_objects["function_program"]
        bs_id = obj_extarq._get_bitstream_id()
        fun_id = obj_extarq._get_function_id()
        md = obj_extarq.get_bitstream_md(
            self.context, fake_dep, fun_id, bs_id)
        self.assertDictEqual(self.images_md["/images"][0], md)

        # bitstream_id is require
        obj_extarq = self.class_fgpa_objects["bitstream_program"]
        bs_id = obj_extarq._get_bitstream_id()
        fun_id = obj_extarq._get_function_id()
        md = obj_extarq.get_bitstream_md(
            self.context, fake_dep, fun_id, bs_id)
        self.assertDictEqual(self.images_md["/images"][0], md)

        # Both bitstream_id and function_id are require
        obj_extarq = self.class_fgpa_objects["bad_program"]
        bs_id = obj_extarq._get_bitstream_id()
        fun_id = obj_extarq._get_function_id()
        self.assertRaises(
            exception.InvalidParameterValue, obj_extarq.get_bitstream_md,
            self.context, fake_dep, fun_id, bs_id)

        # None of bitstream_id or function_id is require
        obj_extarq = self.class_fgpa_objects["no_program"]
        bs_id = obj_extarq._get_bitstream_id()
        fun_id = obj_extarq._get_function_id()
        md = obj_extarq.get_bitstream_md(
            self.context, fake_dep, fun_id, bs_id)
        self.assertIsNone(md)

    @mock.patch('openstack.connection.Connection')
    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    def test_need_extra_bind_job(self, mock_check_state, mock_conn):
        mock_conn.return_value = type(
            "Connection", (object,),
            {"image": type("image", (object,), {"get": self.images_get})})
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(
            self.context, uuid=dep_uuid)

        # function_id is require
        obj_extarq = self.class_fgpa_objects["function_program"]
        need = obj_extarq._need_extra_bind_job(self.context, fake_dep)
        self.assertTrue(need)

        # bitstream_id is require
        obj_extarq = self.class_fgpa_objects["bitstream_program"]
        need = obj_extarq._need_extra_bind_job(self.context, fake_dep)
        self.assertTrue(need)

        # Both bitstream_id and function_id are require
        obj_extarq = self.class_fgpa_objects["bad_program"]
        self.assertRaises(
            exception.InvalidParameterValue, obj_extarq._need_extra_bind_job,
            self.context, fake_dep)

        # None of bitstream_id or function_id is require
        obj_extarq = self.class_fgpa_objects["no_program"]
        need = obj_extarq._need_extra_bind_job(self.context, fake_dep)
        self.assertFalse(need)

    @mock.patch('cyborg.agent.rpcapi.AgentAPI.fpga_program')
    @mock.patch('openstack.connection.Connection')
    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    @mock.patch('cyborg.objects.Deployable.get_cpid_list')
    def test_do_programming(self, mock_cpid_list, mock_check_state, mock_conn,
                            mock_program):
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(
            self.context, uuid=dep_uuid)

        # function_id is require
        obj_extarq = self.class_fgpa_objects["function_program"]
        obj_extarq.arq.hostname = 'newtestnode1'
        fake_dep.driver_name = "intel_fpga"

        mock_cpid_list.return_value = [self.cpid]
        bs_id = obj_extarq._get_bitstream_id()
        obj_extarq._do_programming(self.context, fake_dep, bs_id)
        mock_program.assert_called_once_with(
            self.context, 'newtestnode1', self.cpid, bs_id, "intel_fpga")

    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    @mock.patch('cyborg.objects.Deployable.get_cpid_list')
    def test_do_programming_with_not_one_cp(self, mock_cpid_list,
                                            mock_check_state):
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(
            self.context, uuid=dep_uuid)

        mock_cpid_list.return_value = []
        obj_extarq = self.class_fgpa_objects["function_program"]
        obj_extarq.arq.hostname = 'newtestnode1'
        fake_dep.driver_name = "intel_fpga"
        bs_id = obj_extarq._get_bitstream_id()
        self.assertRaises(exception.ExpectedOneObject,
                          obj_extarq._do_programming,
                          self.context,
                          fake_dep,
                          bs_id)

    @mock.patch('cyborg.common.placement_client.PlacementClient.'
                '__init__')
    @mock.patch('cyborg.common.placement_client.PlacementClient.'
                'add_traits_to_rp')
    @mock.patch('cyborg.common.placement_client.PlacementClient.'
                'delete_traits_with_prefixes')
    def test_update_placement(
        self, mock_delete_traits, mock_add_traits, mock_placement_init):
        mock_placement_init.return_value = None
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(
            self.context, uuid=dep_uuid)

        # function_id is require
        obj_extarq = self.class_fgpa_objects["function_program"]
        obj_extarq.arq.hostname = 'newtestnode1'
        # fake_dep.driver_name = "intel_fpga"
        fun_id = obj_extarq._get_function_id()
        rp_uuid = obj_extarq.arq.device_rp_uuid
        obj_extarq._update_placement(
            self.context, fun_id, fake_dep.driver_name)

        function_id = fun_id.upper().replace('-', '_-')
        vendor = fake_dep.driver_name.upper()
        trait_names = ["_".join((
            constants.FPGA_FUNCTION_ID, vendor, function_id))]
        mock_add_traits.assert_called_once_with(rp_uuid, trait_names)
        mock_delete_traits.assert_called_once_with(
            self.context, rp_uuid, [constants.FPGA_FUNCTION_ID])

    @mock.patch('cyborg.agent.rpcapi.AgentAPI.fpga_program')
    @mock.patch('cyborg.objects.Deployable.get_cpid_list')
    @mock.patch('cyborg.objects.Deployable.update')
    @mock.patch('cyborg.common.placement_client.PlacementClient.'
                '__init__')
    @mock.patch('cyborg.common.placement_client.PlacementClient.'
                'add_traits_to_rp')
    @mock.patch('cyborg.common.placement_client.PlacementClient.'
                'delete_traits_with_prefixes')
    @mock.patch('openstack.connection.Connection')
    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    @mock.patch('cyborg.objects.ExtARQ.bind')
    def test_bind(
        self, mock_bind, mock_check_state, mock_conn, mock_delete_traits,
        mock_add_traits, mock_placement_init, mock_dp_update,
        mock_cpid_list, mock_program):

        mock_placement_init.return_value = None
        mock_conn.return_value = type(
            "Connection", (object,),
            {"image": type("image", (object,), {"get": self.images_get})})

        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(
            self.context, uuid=dep_uuid)

        mock_cpid_list.return_value = [self.cpid]

        # None of bitstream_id or function_id is require
        obj_extarq = self.class_fgpa_objects["no_program"]
        obj_extarq.bind(self.context, fake_dep)
        self.assertFalse(mock_program.called)
        self.assertFalse(mock_placement_init.called)
        self.assertFalse(mock_dp_update.called)
        mock_bind.assert_called_once_with(self.context, fake_dep)

        self.cpid["cpid_info"] = json.dumps(self.bdf).encode('utf-8')
        # function_id is require
        obj_extarq = self.class_fgpa_objects["function_program"]
        obj_extarq.bind(self.context, fake_dep)
        mock_bind.assert_called_with(self.context, fake_dep)
        self.assertEqual(mock_bind.call_count, 2)

        self.cpid["cpid_info"] = json.dumps(self.bdf).encode('utf-8')
        # bitstream_id is require
        obj_extarq = self.class_fgpa_objects["bitstream_program"]
        obj_extarq.bind(self.context, fake_dep)
        mock_bind.assert_called_with(self.context, fake_dep)
        self.assertEqual(mock_bind.call_count, 3)



# Copyright 2020 ZTE Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import itertools

from cyborg.common import policy as old_policy
from cyborg.policies import base
from cyborg.policies import device_profiles


def list_policies():
    return itertools.chain(
        base.list_policies(),
        device_profiles.list_policies(),
        # NOTE(yumeng)old_policies will also be loaded before they are replaced
        # by new policies
        old_policy.device_policies,
        old_policy.deployable_policies,
        old_policy.attribute_policies,
        old_policy.accelerator_request_policies,
        old_policy.fpga_policies,
    )



# Copyright 2019 Intel Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Different accelerator handlers for conductor/agent/api/object to call.
"""


from openstack import connection
from oslo_log import log as logging
from oslo_serialization import jsonutils


from cyborg.agent.rpcapi import AgentAPI
from cyborg.common import constants
from cyborg.common import exception
from cyborg.common import placement_client
from cyborg.common import utils
from cyborg.objects import base
from cyborg.objects.ext_arq import ExtARQ


LOG = logging.getLogger(__name__)


@utils.factory_register(ExtARQ, constants.FPGA)
@base.CyborgObjectRegistry.register
class FPGAExtARQ(ExtARQ):
    """FPGA Extra ARQ."""

    def _get_bitstream_id(self):
        bitstream_id = self.device_profile_group.get(
            constants.ACCEL_BITSTREAM_ID)
        return bitstream_id

    def _get_function_id(self):
        function_id = self.device_profile_group.get(
            constants.ACCEL_FUNCTION_ID)
        return function_id

    def _get_bitstream_md_from_bitstream_id(self, bitstream_id):
        """Get bitstream metadata given a bitstream id."""
        conn = connection.Connection(cloud='devstack-admin')
        resp = conn.image.get('/images/' + bitstream_id)
        if resp:
            return resp.json()
        else:
            LOG.warning('Failed to get image for bitstream (%s)',
                        bitstream_id)
            return None

    # TODO(Shaohe) should move to spec handler.
    def _get_bitstream_md_from_function_id(self, function_id):
        """Get bitstream metadata given a function id."""
        # TODO(Shaohe) parametrize this role in config file.
        conn = connection.Connection(cloud='devstack-admin')
        properties = {constants.ACCEL_FUNCTION_ID: function_id}
        resp = conn.image.get('/images', params=properties)
        if resp:
            image_list = resp.json()['images']
            if type(image_list) != list:
                raise exception.InvalidType(
                    obj='image', type=type(image_list),
                    expected='list')
            if len(image_list) != 1:
                raise exception.ExpectedOneObject(obj='image',
                                                  count=len(image_list))
            LOG.info('[arqs:objs] For function id (%s), got '
                     'bitstream id (%s)', function_id,
                     image_list[0]['id'])
            return image_list[0]
        else:
            LOG.warning('Failed to get image for function (%s)',
                        function_id)
            return None

    def _needs_programming(self, context, deployable):
        bs_id = self._get_bitstream_id()
        fun_id = self._get_function_id()
        if all([bs_id, fun_id]):
            self.update_check_state(
                context, constants.ARQ_BIND_FAILED)
            raise exception.InvalidParameterValue(
                'In device profile {0}, only one among bitstream_id '
                'and function_id must be set, but both are set')
        # TODO(Shaohe) Optimize: check if deployable already has
        # bitstream/function
        if any([bs_id, fun_id]):
            LOG.info('[arqs:objs] bind. Programming needed. '
                     'bitstream: (%s) function: (%s) Deployable UUID: (%s)',
                     bs_id or '', fun_id or '', deployable.uuid)
        else:
            # One situation is that fun_id is zero and device_profile
            # has't bitstream. We should return False.
            LOG.info('No programming is required. ')
            return False
        if bs_id and deployable.bitstream_id == bs_id:
            LOG.info('Deployable %s already has the needed '
                     'bitstream %s. Skipping programming.',
                     deployable.uuid, bs_id)
            return False

        return True

    def get_bitstream_md(self, context, deployable, function_id, bitstream_id):
        """Get bitstream metadata from FPGA image."""
        LOG.info("Get bitstream metadata for deployable(uuid:%s).",
                 deployable.uuid)
        # TODO(Shaohe) Check that deployable.device.hostname matches param
        # hostname out of here
        if not self._needs_programming(context, deployable):
            return

        # FPGA aaS or accelerated Function aaS
        bitstream_md = (
            self._get_bitstream_md_from_bitstream_id(bitstream_id)
            if bitstream_id else
            self._get_bitstream_md_from_function_id(function_id))
        if bitstream_md:
            LOG.info('ARQ %s get bitstream metadata:%s from image registry.',
                     self.arq.uuid, bitstream_md)
        else:
            self.update_check_state(
                context, constants.ARQ_BIND_FAILED)
            LOG.error('Can not get bitstream metadata from image registry '
                      'for ARQ %s', self.arq.uuid)
        return bitstream_md

    def _need_extra_bind_job(self, context, deployable):
        return self._needs_programming(context, deployable)

    @utils.wrap_job_tb("Error during ARQ bind job. Reason: %s")
    def bind(self, context, deployable):
        LOG.info('Start bind jobs for ARQ(%s) with deployable(%s)',
                 self.arq.uuid, deployable.uuid)
        bs_id = self._get_bitstream_id()
        fun_id = self._get_function_id()
        bs_md = self.get_bitstream_md(context, deployable, fun_id, bs_id)
        ok = False
        if bs_md:
            ok = self._do_programming(context, deployable, bs_md['id'])
        if ok:
            fun_id = fun_id or bs_md[constants.ACCEL_FUNCTION_ID]
            self._update_placement(context, fun_id, deployable.driver_name)
            deployable.update(context, {'bitstream_id': bs_md['id']})

        super(FPGAExtARQ, self).bind(context, deployable)

        return True

    def _unbind(self):
        # TODO(Shaohe) add cancel _update_placement, unbind operation.
        pass

    def _delete(self):
        # TODO(Shaohe) add cancel _update_placement, delete operation.
        pass

    def _update_placement(self, context, function_id, driver_name):
        """update resources provider after program."""
        # TODO(Sundar) Don't apply function trait if bitstream is private
        if not function_id:
            LOG.info("Not get function id for resources provider %s.",
                     self.arq.device_rp_uuid)
            return

        placement = placement_client.PlacementClient()
        try:
            placement.delete_traits_with_prefixes(
                context, self.arq.device_rp_uuid,
                [constants.FPGA_FUNCTION_ID])
        except Exception as e:
            LOG.error("Failed to delete traits(%s) from resources provider %s."
                      "Reason: %s", constants.FPGA_FUNCTION_ID,
                      self.arq.device_rp_uuid, e.message)
            self.update_check_state(
                context, constants.ARQ_BIND_FAILED)
            raise

        function_id = function_id.upper().replace('-', '_-')
        # TODO(Sundar) Validate this is a valid trait name
        vendor = driver_name.upper()
        trait_names = ["_".join((
            constants.FPGA_FUNCTION_ID, vendor, function_id))]
        try:
            placement.add_traits_to_rp(
                self.arq.device_rp_uuid, trait_names)
        except Exception as e:
            LOG.error("Failed to add traits(%s) to resources provider %s."
                      "Reason: %s", trait_names,
                      self.arq.device_rp_uuid, e.message)
            # TODO(Shaohe) Rollback? We have _update_placement,
            # should cancel it.
            self.update_check_state(
                context, constants.ARQ_BIND_FAILED)
            raise
        LOG.info("Add traits(%s) to resources provider %s.",
                 trait_names, self.arq.device_rp_uuid)

    def _do_programming(self, context, deployable, bitstream_id):
        """FPGA program."""
        hostname = self.arq.hostname
        driver_name = deployable.driver_name

        # query_filter = {"device_id": deployable.device_id}
        # TODO(Shaohe) We should probably get cpid from objects layer,
        # not db layer
        cpid_list = deployable.get_cpid_list(context)
        count = len(cpid_list)
        if count != 1:
            self.update_check_state(
                context, constants.ARQ_BIND_FAILED)
            raise exception.ExpectedOneObject(obj='controlpath_id',
                                              count=count)
        controlpath_id = cpid_list[0]
        controlpath_id['cpid_info'] = jsonutils.loads(
            controlpath_id['cpid_info'])
        LOG.info('Found control path id: %s', controlpath_id)

        LOG.info('Starting programming for host: (%s) deployable (%s) '
                 'bitstream_id (%s)', hostname,
                 deployable.uuid, bitstream_id)
        # TODO(Shaohe) do this asynchronously, do this in conductor or agent?
        try:
            agent = AgentAPI()
            agent.fpga_program(context, hostname,
                               controlpath_id, bitstream_id,
                               driver_name)
        except Exception as e:
            self.update_check_state(
                context, constants.ARQ_BIND_FAILED)
            LOG.error('Failed programming for host: (%s) deployable (%s). '
                      'Error: %s', hostname, deployable.uuid, e.message)
            raise
        LOG.info('Finished programming for host: (%s) deployable (%s)',
                 hostname, deployable.uuid)
        # TODO(Shaohe) propagate agent errors to caller
        return True



# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
import pecan
import wsme
from wsme import types as wtypes

from oslo_log import log

from cyborg.api.controllers import base
from cyborg.api.controllers import link
from cyborg.api.controllers import types
from cyborg.api.controllers.v2 import utils
from cyborg.api.controllers.v2 import versions
from cyborg.api import expose
from cyborg.common import authorize_wsgi
from cyborg.common import constants
from cyborg.common import exception
from cyborg.common.i18n import _
from cyborg import objects

LOG = log.getLogger(__name__)


class ARQ(base.APIBase):
    """API representation of an ARQ.

    This class enforces type checking and value constraints, and converts
    between the internal object model and the API representation.
    """
    uuid = types.uuid
    """The UUID of the ARQ"""

    state = wtypes.text  # obvious meanings
    device_profile_name = wtypes.text
    device_profile_group_id = wtypes.IntegerType()

    hostname = wtypes.text
    """The host name to which the ARQ is bound, if any"""

    device_rp_uuid = types.uuid
    """The UUID of the bound device RP, if any"""

    instance_uuid = types.uuid
    """The UUID of the instance associated with this ARQ, if any"""
    project_id = types.uuid
    """The UUID of the instance project_id associated with this ARQ, if any"""

    attach_handle_type = wtypes.text
    attach_handle_uuid = wtypes.text
    attach_handle_info = {wtypes.text: wtypes.text}

    links = wsme.wsattr([link.Link], readonly=True)
    """A list containing a self link"""

    def __init__(self, **kwargs):
        super(ARQ, self).__init__(**kwargs)
        self.fields = []
        for field in objects.ARQ.fields:
            self.fields.append(field)
            setattr(self, field, kwargs.get(field, wtypes.Unset))

    @classmethod
    def convert_with_links(cls, obj_arq):
        api_arq = cls(**obj_arq.as_dict())
        api_arq.links = [
            link.Link.make_link('self', pecan.request.public_url,
                                'accelerator_requests', api_arq.uuid)
            ]
        return api_arq


class ARQCollection(base.APIBase):
    """API representation of a collection of arqs."""

    arqs = [ARQ]
    """A list containing arq objects"""

    @classmethod
    def convert_with_links(cls, obj_arqs):
        collection = cls()
        collection.arqs = [ARQ.convert_with_links(obj_arq)
                           for obj_arq in obj_arqs]
        return collection


class ARQsController(base.CyborgController):
    """REST controller for ARQs.

       For the relationship between ARQs and device profiles, see
       nova/nova/accelerator/cyborg.py.
    """

    @authorize_wsgi.authorize_wsgi("cyborg:arq", "create", False)
    @expose.expose(ARQCollection, body=types.jsontype,
                   status_code=HTTPStatus.CREATED)
    def post(self, req):
        """Create one or more ARQs for a single device profile.
           Request body:
              { 'device_profile_name': <string> }
           Future:
              { 'device_profile_name': <string> # required
                'device_profile_group_id': <integer>, # opt, default=0
                'image_uuid': <glance-image-UUID>, #optional, for future
              }
           :param req: request body.
        """
        LOG.info("[arq] post req = (%s)", req)
        context = pecan.request.context
        dp_name = req.get('device_profile_name')
        if dp_name is not None:
            try:
                devprof = objects.DeviceProfile.get_by_name(context, dp_name)
            except exception.ResourceNotFound:
                raise exception.ResourceNotFound(
                    resource='Device Profile',
                    msg='with name=%s' % dp_name)
            except Exception as e:
                raise e
        else:
            raise exception.DeviceProfileNameNeeded()
        LOG.info('[arqs] post. device profile name=%s', dp_name)

        extarq_list = []
        for group_id, group in enumerate(devprof.groups):
            accel_resources = []
            # If the device profile requires the Xilinx fpga, the number of
            # resources should multiply by 2 cause that end user can program
            # the device only when both MGMT and USER PF are bound to
            # instance.
            if group.get("trait:CUSTOM_FPGA_XILINX") == "required":
                accel_resources = [int(group.get("resources:FPGA"))] * 2
            else:
                accel_resources = [
                    int(val) for key, val in group.items()
                    if key.startswith('resources')]

            # If/when we introduce non-accelerator resources, like
            # device-local memory, the key search above needs to be
            # made specific to accelerator resources only.
            num_accels = sum(accel_resources)
            arq_fields = {
                'device_profile_name': devprof.name,
                'device_profile_group_id': group_id,
            }
            for i in range(num_accels):
                obj_arq = objects.ARQ(context, **arq_fields)
                extarq_fields = {'arq': obj_arq}
                obj_extarq = objects.ExtARQ(context, **extarq_fields)
                new_extarq = pecan.request.conductor_api.arq_create(
                    context, obj_extarq, devprof.id)
                extarq_list.append(new_extarq)

        ret = ARQCollection.convert_with_links(
            [extarq.arq for extarq in extarq_list])
        LOG.info('[arqs] post returned: %s', ret)
        return ret

    @authorize_wsgi.authorize_wsgi("cyborg:arq", "get_one")
    @expose.expose(ARQ, types.uuid)
    def get_one(self, uuid):
        """Get a single ARQ by UUID."""
        context = pecan.request.context
        extarq = objects.ExtARQ.get(context, uuid)
        return ARQ.convert_with_links(extarq.arq)

    @authorize_wsgi.authorize_wsgi("cyborg:arq", "get_all", False)
    @expose.expose(ARQCollection, wtypes.text, types.uuid)
    def get_all(self, bind_state=None, instance=None):
        """Retrieve a list of arqs."""
        # TODO(Sundar) Need to implement 'arq=uuid1,...' query parameter
        LOG.info('[arqs] get_all. bind_state:(%s), instance:(%s)',
                 bind_state or '', instance or '')
        context = pecan.request.context
        extarqs = objects.ExtARQ.list(context)
        state_map = constants.ARQ_BIND_STATES_STATUS_MAP
        valid_bind_states = list(state_map.keys())
        arqs = [extarq.arq for extarq in extarqs]
        # TODO(Sundar): Optimize by doing the filtering in the db layer
        # Apply instance filter before state filter.
        if bind_state and bind_state != 'resolved':
            raise exception.ARQBadState(
                state=bind_state, uuid=None, expected=['resolved'])
        if instance:
            new_arqs = [arq for arq in arqs
                        if arq['instance_uuid'] == instance]
            arqs = new_arqs
            if bind_state:
                for arq in new_arqs:
                    if arq['state'] not in valid_bind_states:
                        # NOTE(Sundar) This should return HTTP code 423
                        # if any ARQ for this instance is not resolved.
                        LOG.warning('Some of ARQs for instance %s is not '
                                    'resolved', instance)
                        return wsme.api.Response(
                            None,
                            status_code=HTTPStatus.LOCKED)
        elif bind_state:
            arqs = [arq for arq in arqs
                    if arq['state'] in valid_bind_states]

        ret = ARQCollection.convert_with_links(arqs)
        LOG.info('[arqs:get_all] Returned: %s', ret)
        return ret

    @authorize_wsgi.authorize_wsgi("cyborg:arq", "delete", False)
    @expose.expose(None, wtypes.text, types.uuid,
                   status_code=HTTPStatus.NO_CONTENT)
    def delete(self, arqs=None, instance=None):
        """Delete one or more ARQS.

        The request can be either one of these two forms:
            DELETE /v2/accelerator_requests?arqs=uuid1,uuid2,...
            DELETE /v2/accelerator_requests?instance=uuid

        The second form is idempotent, i.e., it would have the same effect
        if called repeatedly with the same instance UUID. In other words,
        it would not raise an error on the second and later attempts even if
        the first one has deleted the ARQs. Whereas the first form is not
        idempotent: if one or more of the ARQs do not exist, it would raise
        an error. Nova uses the second form: so repeated calls do not cause
        issues.

        :param arqs: List of ARQ UUIDs
        :param instance: UUID of instance whose ARQs need to be deleted
        """
        context = pecan.request.context
        if (arqs and instance) or (not arqs and not instance):
            raise exception.ObjectActionError(
                action='delete',
                reason='Provide either an ARQ uuid list or an instance UUID')
        elif arqs:
            LOG.info("[arqs] delete. arqs=(%s)", arqs)
            pecan.request.conductor_api.arq_delete_by_uuid(context, arqs)
        else:  # instance is not None
            LOG.info("[arqs] delete. instance=(%s)", instance)
            pecan.request.conductor_api.arq_delete_by_instance_uuid(
                context, instance)

    def _validate_arq_patch(self, patch):
        """Validate a single patch for an ARQ.

        :param patch: a JSON PATCH document.
            The patch must be of the form [{..}], as specified in the
            value field of arq_uuid in patch() method below.
        :returns: dict of valid fields
        """
        valid_fields = {'hostname': None,
                        'device_rp_uuid': None,
                        'instance_uuid': None}
        if utils.allow_project_id():
            valid_fields['project_id'] = None
        if ((not all(p['op'] == 'add' for p in patch)) and
           (not all(p['op'] == 'remove' for p in patch))):
            raise exception.PatchError(
                reason='Every op must be add or remove')

        for p in patch:
            path = p['path'].lstrip('/')
            if path == 'project_id' and not utils.allow_project_id():
                raise exception.NotAcceptable(_(
                    "Request not acceptable. The minimal required API "
                    "version should be %(base)s.%(opr)s") %
                    {'base': versions.BASE_VERSION,
                     'opr': versions.MINOR_1_PROJECT_ID})
            if path not in valid_fields.keys():
                reason = 'Invalid path in patch {}'.format(p['path'])
                raise exception.PatchError(reason=reason)
            if p['op'] == 'add':
                valid_fields[path] = p['value']
        not_found = [field for field, value in valid_fields.items()
                     if value is None]
        if patch[0]['op'] == 'add' and len(not_found) > 0:
            msg = ','.join(not_found)
            reason = _('Fields absent in patch {}').format(msg)
            raise exception.PatchError(reason=reason)

        return valid_fields

    @staticmethod
    def _check_if_already_bound(context, valid_fields):
        patch_fields = list(valid_fields.values())[0]
        instance_uuid = patch_fields['instance_uuid']
        extarqs = objects.ExtARQ.list(context)
        extarqs_for_instance = [
            extarq for extarq in extarqs
            if extarq.arq['instance_uuid'] == instance_uuid]
        if extarqs_for_instance:  # duplicate binding request
            msg = _('Instance {} already has accelerator requests. '
                    'Cannot bind additional ARQs.')
            reason = msg.format(instance_uuid)
            raise exception.PatchError(reason=reason)

    @authorize_wsgi.authorize_wsgi("cyborg:arq", "update", False)
    @expose.expose(None, body=types.jsontype,
                   status_code=HTTPStatus.ACCEPTED)
    def patch(self, patch_list):
        """Bind/Unbind one or more ARQs.

        Usage: curl -X PATCH .../v2/accelerator_requests
                 -d <patch_list> -H "Content-type: application/json"

        :param patch_list: A map from ARQ UUIDs to their JSON patches:
            {"$arq_uuid": [
                {"path": "/hostname", "op": ADD/RM, "value": "..."},
                {"path": "/device_rp_uuid", "op": ADD/RM, "value": "..."},
                {"path": "/instance_uuid", "op": ADD/RM, "value": "..."},
                {"path": "/project_id", "op": ADD/RM, "value": "..."},
               ],
             "$arq_uuid": [...]
            }
            In particular, all and only these 4 fields must be present,
            and only 'add' or 'remove' ops are allowed.
        """
        LOG.info('[arqs] patch. list=(%s)', patch_list)
        context = pecan.request.context
        # Validate all patches before un/binding.
        valid_fields = {}
        for arq_uuid, patch in patch_list.items():
            valid_fields[arq_uuid] = self._validate_arq_patch(patch)

        # NOTE(Sundar): In the ARQ create/bind flow, new ARQs can be created
        # for a device profile any time. However, they should not be bound to
        # an instance which already has other ARQs bound to it. In the future,
        # we may allow that for hot adds, but not now.
        # See commit message of https://review.opendev.org/712231 for details.
        #
        # So, for bind requests, we first check that no ARQs are already
        # associated with the instance specified in the binding.
        patch = list(patch_list.values())[0]
        if patch[0]['op'] == 'add':
            self._check_if_already_bound(context, valid_fields)

        pecan.request.conductor_api.arq_apply_patch(
            context, patch_list, valid_fields)



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as config_fixture
from oslo_context import context
from oslo_db import options
from oslo_log import log
from oslo_utils import excutils
from oslotest import base

import contextlib
import eventlet
import testtools

from cyborg.common import config as cyborg_config
from cyborg.tests import post_mortem_debug
from cyborg.tests.unit import policy_fixture


CONF = cfg.CONF
options.set_defaults(cfg.CONF)
try:
    log.register_options(CONF)
except cfg.ArgsAlreadyParsedError:
    pass


class TestCase(base.BaseTestCase):
    """Test case base class for all unit tests."""

    def setUp(self):
        super(TestCase, self).setUp()
        self.context = context.get_admin_context()
        self._set_config()
        self.policy = self.useFixture(policy_fixture.PolicyFixture())

    def _set_config(self):
        self.cfg_fixture = self.useFixture(config_fixture.Config(cfg.CONF))
        self.config(use_stderr=False,
                    fatal_exception_format_errors=True)
        self.set_defaults(host='fake-mini',
                          debug=True)
        self.set_defaults(connection="sqlite://",
                          sqlite_synchronous=False,
                          group='database')
        cyborg_config.parse_args([], default_config_files=[])

    def config(self, **kw):
        """Override config options for a test."""
        self.cfg_fixture.config(**kw)

    def set_defaults(self, **kw):
        """Set default values of config options."""
        group = kw.pop('group', None)
        for o, v in kw.items():
            self.cfg_fixture.set_default(o, v, group=group)

    def get_path(self, project_file=None):
        """Get the absolute path to a file. Used for testing the API.

        :param project_file: File whose path to return. Default: None.
        :returns: path to the specified file, or path to project root.
        """
        root = os.path.abspath(
            os.path.join(os.path.dirname(__file__), '..', '..')
            )
        if project_file:
            return os.path.join(root, project_file)
        else:
            return root


# Test worker cannot survive eventlet's Timeout exception, which effectively
# kills the whole worker, with all test cases scheduled to it. This metaclass
# makes all test cases convert Timeout exceptions into unittest friendly
# failure mode (self.fail).
class DietTestCase(base.BaseTestCase):
    """Same great taste, less filling.

    BaseTestCase is responsible for doing lots of plugin-centric setup
    that not all tests require (or can tolerate).  This class provides
    only functionality that is common across all tests.
    """

    def setUp(self):
        super(DietTestCase, self).setUp()
        self.context = context.get_admin_context()

        options.set_defaults(cfg.CONF, connection='sqlite://')

        debugger = os.environ.get('OS_POST_MORTEM_DEBUGGER')
        if debugger:
            self.addOnException(post_mortem_debug.get_exception_handler(
                debugger))

        self.addCleanup(mock.patch.stopall)

        self.addOnException(self.check_for_systemexit)
        self.orig_pid = os.getpid()

    def addOnException(self, handler):

        def safe_handler(*args, **kwargs):
            try:
                return handler(*args, **kwargs)
            except Exception:
                with excutils.save_and_reraise_exception(reraise=False) as ctx:
                    self.addDetail('Failure in exception handler %s' % handler,
                                   testtools.content.TracebackContent(
                                       (ctx.type_, ctx.value, ctx.tb), self))

        return super(DietTestCase, self).addOnException(safe_handler)

    def check_for_systemexit(self, exc_info):
        if isinstance(exc_info[1], SystemExit):
            if os.getpid() != self.orig_pid:
                # Subprocess - let it just exit
                raise
            # This makes sys.exit(0) still a failure
            self.force_failure = True

    @contextlib.contextmanager
    def assert_max_execution_time(self, max_execution_time=5):
        with eventlet.Timeout(max_execution_time, False):
            yield
            return
        self.fail('Execution of this test timed out')

    def assertOrderedEqual(self, expected, actual):
        expect_val = self.sort_dict_lists(expected)
        actual_val = self.sort_dict_lists(actual)
        self.assertEqual(expect_val, actual_val)

    def sort_dict_lists(self, dic):
        for key, value in dic.items():
            if isinstance(value, list):
                dic[key] = sorted(value)
            elif isinstance(value, dict):
                dic[key] = self.sort_dict_lists(value)
        return dic

    def assertDictSupersetOf(self, expected_subset, actual_superset):
        """Checks that actual dict contains the expected dict.

        After checking that the arguments are of the right type, this checks
        that each item in expected_subset is in, and matches, what is in
        actual_superset. Separate tests are done, so that detailed info can
        be reported upon failure.
        """
        if not isinstance(expected_subset, dict):
            self.fail("expected_subset (%s) is not an instance of dict" %
                      type(expected_subset))
        if not isinstance(actual_superset, dict):
            self.fail("actual_superset (%s) is not an instance of dict" %
                      type(actual_superset))
        for k, v in expected_subset.items():
            self.assertIn(k, actual_superset)
            self.assertEqual(v, actual_superset[k],
                             "Key %(key)s expected: %(exp)r, actual %(act)r" %
                             {'key': k, 'exp': v, 'act': actual_superset[k]})



# Copyright 2020 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg Intel NIC driver implementation.
"""

from cyborg.accelerator.drivers.nic.base import NICDriver
from cyborg.accelerator.drivers.nic.intel import sysinfo


class IntelNICDriver(NICDriver):
    """Class for Intel NIC drivers.
       Vendor should implement their specific drivers in this class.
    """
    VENDOR = "intel"

    def __init__(self, *args, **kwargs):
        pass

    def discover(self):
        return sysinfo.nic_tree()



# Copyright 2018 Lenovo (Beijing) Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import base as object_base

from cyborg.objects.attribute import Attribute
from cyborg.objects import base
from cyborg.objects import fields as object_fields


@base.CyborgObjectRegistry.register
class DriverAttribute(base.DriverObjectBase,
                      object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'key': object_fields.StringField(nullable=False),
        'value': object_fields.StringField(nullable=False)
    }

    def create(self, context, deployable_id):
        """Convert driver-side Attribute into Attribute Object so as to
        store in DB.
        """
        attr_obj = Attribute()
        attr_obj.deployable_id = deployable_id
        attr_obj.set_key_value_pair(self.key, self.value)
        attr_obj.create(context)

    @classmethod
    def destroy(cls, context, deployable_id):
        """Delete driver-side attribute list from the DB."""
        attr_obj_list = Attribute.get_by_deployable_id(context, deployable_id)
        for attr_obj in attr_obj_list:
            attr_obj.destroy(context)

    @classmethod
    def delete_by_key(cls, context, deployable_id, key):
        """Delete driver-side attribute list from the DB."""
        attr_obj_list = Attribute.get_by_deployable_id(context, deployable_id)
        for attr_obj in attr_obj_list:
            if key == attr_obj.key:
                attr_obj.destroy(context)

    @classmethod
    def list(cls, context, deployable_id):
        """Form driver-side attribute list for one deployable."""
        attr_obj_list = Attribute.get_by_deployable_id(context, deployable_id)
        driver_attr_obj_list = []
        for attr_obj in attr_obj_list:
            driver_attr_obj = cls(context=context,
                                  key=attr_obj.key,
                                  value=attr_obj.value)
            driver_attr_obj_list.append(driver_attr_obj)
        return driver_attr_obj_list



# Copyright 2020 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

pci_group = cfg.OptGroup(
    name='pci',
    title='PCI passthrough options')

pci_opts = [
    cfg.MultiStrOpt('passthrough_whitelist',
                    default=[],
                    help=" ")
]

nic_group = cfg.OptGroup(
    name='nic_devices',
    title='nic device ID options',
    help="""This is used to config specific nic devices.
    """)

nic_opts = [
    cfg.ListOpt('enabled_nic_types',
                default=[],
                help=" ")
]

gpu_group = cfg.OptGroup(
    name='gpu_devices',
    title='virtual gpu options',
    help="""This is used to config vGPU types for nvidia GPU devices.
    """)

vgpu_opts = [
    cfg.ListOpt('enabled_vgpu_types',
                default=[],
                help="""
The vGPU types enabled in the compute node.

Cyborg supports multiple vGPU types in one host. Usually, a single physical
GPU can only set one vgpu type. Some pGPUs (e.g. NVIDIA GRID K1) support
multiple vGPU types.

If more than one single vGPU type are provided, then for each
*vGPU type*, you must add an additional section ``[vgpu_$(VGPU_TYPE)]`` with
a single configuration option ``device_addresses`` to assign this type to
the target physical GPU(s). PGPUs should be configured explictly now, we will
improve this after we implement the enable/disable interface.

If the same PCI address is provided for two different types, cyborg-agent will
return an InvalidGPUConfig exception at restart.

An example is as the following::

    [gpu_devices]
    enabled_vgpu_types = nvidia-35, nvidia-36

    [vgpu_nvidia-35]
    device_addresses = 0000:84:00.0,0000:85:00.0

    [vgpu_nvidia-36]
    device_addresses = 0000:86:00.0

""")
]


def register_opts(conf):
    conf.register_group(nic_group)
    conf.register_opts(nic_opts, group=nic_group)
    conf.register_group(gpu_group)
    conf.register_opts(vgpu_opts, group=gpu_group)
    conf.register_group(pci_group)
    conf.register_opts(pci_opts, group=pci_group)


def register_dynamic_opts(conf):
    """Register dynamically-generated options and groups.

    This must be called by the service that wishes to use the options **after**
    the initial configuration has been loaded.
    """
    opts = [
        cfg.ListOpt('physical_device_mappings', default=[],
                    item_type=cfg.types.String()),
        cfg.ListOpt('function_device_mappings', default=[],
                    item_type=cfg.types.String()),
    ]

    # Register the '[nic_type]/physical_device_mappings' and
    # '[nic_type]/function_device_mappings' opts, implicitly
    # registering the '[nic_type]' groups in the process
    for nic_type in conf.nic_devices.enabled_nic_types:
        conf.register_opts(opts, group=nic_type)
    # Register the '[vgpu_$(VGPU_TYPE)]/device_addresses' opts, implicitly
    # registering the '[vgpu_$(VGPU_TYPE)]' groups in the process
    opt = cfg.ListOpt('device_addresses', default=[],
                      item_type=cfg.types.String())
    for vgpu_type in conf.gpu_devices.enabled_vgpu_types:
        conf.register_opt(opt, group='vgpu_%s' % vgpu_type)


def list_opts():
    return {nic_group: nic_opts, gpu_group: vgpu_opts}



# Copyright (c) 2018 Intel.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Track resources like FPGA GPU and QAT for a host.  Provides the
conductor with useful information about availability through the accelerator
model.
"""

from oslo_log import log as logging
from stevedore import driver
from stevedore.extension import ExtensionManager

from cyborg.common import exception
from cyborg.common import utils
from cyborg.conf import CONF


LOG = logging.getLogger(__name__)

AGENT_RESOURCE_SEMAPHORE = "agent_resources"


class ResourceTracker(object):
    """Agent helper class for keeping track of resource usage as instances
    are built and destroyed.
    """

    def __init__(self, host, cond_api):
        self.host = host
        self.conductor_api = cond_api
        self.acc_drivers = []
        self._initialize_drivers()

    def _initialize_drivers(self, enabled_drivers=None):
        """Load accelerator drivers.

        :return: [nvidia_gpu_driver_obj, intel_fpga_driver_obj]
        """
        acc_drivers = []
        if not enabled_drivers:
            enabled_drivers = CONF.agent.enabled_drivers
        valid_drivers = ExtensionManager(
            namespace='cyborg.accelerator.driver').names()
        for d in enabled_drivers:
            if d not in valid_drivers:
                raise exception.InvalidDriver(name=d)
            acc_driver = driver.DriverManager(
                namespace='cyborg.accelerator.driver', name=d,
                invoke_on_load=True).driver
            acc_drivers.append(acc_driver)
        self.acc_drivers = acc_drivers

    @utils.synchronized(AGENT_RESOURCE_SEMAPHORE)
    def update_usage(self, context):
        """Update the resource usage periodically.
        """
        acc_list = []
        for acc_driver in self.acc_drivers:
            acc_list.extend(acc_driver.discover())
        # Call conductor_api here to diff and report acc data. Now, we actually
        # do not have the method report_data.
        try:
            self.conductor_api.report_data(context, self.host, acc_list)
        except exception.PlacementResourceProviderNotFound as e:
            LOG.error('Unable to report usage: %s', e)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg

SERVICE_USER_GROUP = 'service_user'

service_user = cfg.OptGroup(
    SERVICE_USER_GROUP,
    title='Service token authentication type options',
    help="""
Configuration options for service to service authentication using a service
token. These options allow sending a service token along with the user's token
when contacting external REST APIs.
"""
)

service_user_opts = [
    cfg.BoolOpt('send_service_user_token',
                default=False,
                help="""
When True, if sending a user token to a REST API, also send a service token.
"""),
]


def register_opts(conf):
    conf.register_group(service_user)
    conf.register_opts(service_user_opts, group=service_user)

    ks_loading.register_session_conf_options(conf, SERVICE_USER_GROUP)
    ks_loading.register_auth_conf_options(conf, SERVICE_USER_GROUP)


def list_opts():
    return {
        service_user: (
            service_user_opts +
            ks_loading.get_session_conf_options() +
            ks_loading.get_auth_common_conf_options() +
            ks_loading.get_auth_plugin_conf_options('password') +
            ks_loading.get_auth_plugin_conf_options('v2password') +
            ks_loading.get_auth_plugin_conf_options('v3password'))
    }



# Modifications Copyright (C) 2021 ZTE Corporation
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Utils for GPU driver.
"""
from oslo_concurrency import processutils
from oslo_log import log as logging

import re

import cyborg.common.exception as exception
import cyborg.conf
import cyborg.privsep

LOG = logging.getLogger(__name__)

GPU_FLAGS = ["VGA compatible controller", "3D controller"]
GPU_INFO_PATTERN = re.compile(r"(?P<devices>[0-9a-fA-F]{4}:[0-9a-fA-F]{2}:"
                              r"[0-9a-fA-F]{2}\.[0-9a-fA-F]) "
                              r"(?P<controller>.*) [\[].*]: (?P<model>.*) .*"
                              r"[\[](?P<vendor_id>[0-9a-fA-F]"
                              r"{4}):(?P<product_id>[0-9a-fA-F]{4})].*")

VENDOR_MAPS = {"10de": "nvidia", "102b": "matrox"}
PRODUCT_ID_MAPS = {"1eb8": "T4", "15f7": "P100_PCIE_12GB"}


@cyborg.privsep.sys_admin_pctxt.entrypoint
def lspci_privileged():
    cmd = ['lspci', '-nnn', '-D']
    return processutils.execute(*cmd)


@cyborg.privsep.sys_admin_pctxt.entrypoint
def create_mdev_privileged(pci_addr, mdev_type, ah_uuid):
    """Instantiate a mediated device."""
    if ah_uuid is None:
        raise exception.AttachHandleUUIDNeeded()
    fpath = '/sys/class/mdev_bus/{0}/mdev_supported_types/{1}/create'
    fpath = fpath.format(pci_addr, mdev_type)
    with open(fpath, 'w') as f:
        f.write(ah_uuid)
    return ah_uuid


@cyborg.privsep.sys_admin_pctxt.entrypoint
def remove_mdev_privileged(physical_device, mdev_type, medv_uuid):
    fpath = ('/sys/class/mdev_bus/{0}/mdev_supported_types/'
             '{1}/devices/{2}/remove')
    fpath = fpath.format(physical_device, mdev_type, medv_uuid)
    with open(fpath, 'w') as f:
        f.write("1")


def get_pci_devices(pci_flags, vendor_id=None):
    device_for_vendor_out = []
    all_device_out = []
    lspci_out = lspci_privileged()[0].split('\n')
    for pci in lspci_out:
        if any(x in pci for x in pci_flags):
            all_device_out.append(pci)
            if vendor_id and vendor_id in pci:
                device_for_vendor_out.append(pci)
    return device_for_vendor_out if vendor_id else all_device_out


def discover_vendors():
    vendors = set()
    gpus = get_pci_devices(GPU_FLAGS)
    for gpu in gpus:
        m = GPU_INFO_PATTERN.match(gpu)
        if m:
            vendor_id = m.groupdict().get("vendor_id")
            vendors.add(vendor_id)
    return vendors



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Cyborg SPDK driver modules implementation.
"""

from oslo_log import log as logging
LOG = logging.getLogger(__name__)


class SPDKDRIVER(object):
    """SPDKDRIVER

        This is just a virtual SPDK drivers interface.
        SPDK-based app server should implement their specific drivers.
    """
    @classmethod
    def create(cls, server, *args, **kwargs):
        for subclass in cls.__subclasses__():
            if server == subclass.SERVER:
                return subclass(*args, **kwargs)
        raise LookupError("Could not find the driver for server %s" % server)

    def __init__(self, *args, **kwargs):
        super(SPDKDRIVER, self).__init__()

    def discover_accelerator(self):
        """Discover a backend accelerator

        :return: accelerator list.
        """
        raise NotImplementedError('Subclasses must implement this method.')

    def install_accelerator(self, driver_id, driver_type):
        """install a backend accelerator

        :param driver_id: driver id.
        :param driver_type: driver type.

        :raise: NotImplementedError.
        """
        raise NotImplementedError('Subclasses must implement this method.')

    def uninstall_accelerator(self, driver_id, driver_type):
        """uninstall a backend accelerator

        :param driver_id: driver id.
        :param driver_type: driver type.

        :raise: NotImplementedError.
        """
        raise NotImplementedError('Subclasses must implement this method.')

    def accelerator_list(self):
        """Discover a backend accelerator list

        :return: accelerator list.
        :raise: NotImplementedError.
        """
        raise NotImplementedError('Subclasses must implement this method.')

    def update(self, driver_type, **kwargs):
        """update

        :param driver_type: driver type.
        :param kwargs: kwargs.
        :raise: NotImplementedError.
        """
        raise NotImplementedError('Subclasses must implement this method.')

    def attach_instance(self, instance_id):
        """attach a backend instance

        :param instance_id: instance id.
        :return: instance.
        :raise: NotImplementedError.
        """
        raise NotImplementedError('Subclasses must implement this method.')

    def detach_instance(self, instance_id):
        """detach a backend instance

        :param instance_id: instance id.
        :return: instance.
        :raise: NotImplementedError.
        """
        raise NotImplementedError('Subclasses must implement this method.')



# Copyright 2018 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cyborg.accelerator.drivers.fpga.base import FPGADriver
from cyborg.accelerator.drivers.fpga.intel.driver import IntelFPGADriver  # noqa
from cyborg.accelerator.drivers.fpga.inspur.driver import InspurFPGADriver  # noqa
from cyborg.accelerator.drivers.fpga.xilinx.driver import XilinxFPGADriver  # noqa
from cyborg.tests import base


class TestFPGADriver(base.TestCase):
    def test_create(self):
        FPGADriver.create("intel")
        FPGADriver.create("inspur")
        FPGADriver.create("xilinx")
        self.assertRaises(LookupError, FPGADriver.create, "fake")

    def test_discover(self):
        d = FPGADriver()
        self.assertRaises(NotImplementedError, d.discover)

    def test_program(self):
        d = FPGADriver()
        self.assertRaises(NotImplementedError, d.program, "path", "image")



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from cyborg.common.i18n import _


opts = [
    cfg.StrOpt('mysql_engine',
               default='InnoDB',
               help=_('MySQL engine to use.'))
]

opt_group = cfg.OptGroup(name='database',
                         title='Options for the database service')


def register_opts(conf):
    conf.register_opts(opts, group=opt_group)


DB_OPTS = (opts)


def list_opts():
    return {
        opt_group: DB_OPTS
    }
